# pylint: disable=R0401
"""
@file
@brief Helpers to compile C.
"""
import os
import sys
import shutil
import numpy


_header_c_float = """
void concat_float_float(float* xy, float x, float y)
{
    xy[0] = x;
    xy[1] = y;
}

void adot_float_float(float* res, float* vx, float* vy, int dim)
{
    *res = 0;
    for(; dim > 0; --dim, ++vx, ++vy)
        *res += *vx * *vy;
}

void aadd_float(float* res, float* vx, float* vy, int dim)
{
    for(; dim > 0; --dim, ++vx, ++vy, ++res)
        *res = *vx + *vy;
}

void asub_float_float(float* res, float* vx, float* vy, int dim)
{
    for(; dim > 0; --dim, ++vx, ++vy, ++res)
        *res = *vx - *vy;
}

void amul_float_float(float* res, float* vx, float* vy, int dim)
{
    for(; dim > 0; --dim, ++vx, ++vy, ++res)
        *res = *vx * *vy;
}

void adiv_float_float(float* res, float* vx, float* vy, int dim)
{
    for(; dim > 0; --dim, ++vx, ++vy, ++res)
        *res = *vx / *vy;
}

void sign_float(float* res, float x)
{
    *res = x >= 0 ? (float)1 : (float)0 ;
}

void atake_float_int(float* res, float * vx, int p, int dim)
{
    *res = vx[p];
}

void atake_int_int(int* res, int* vx, int p, int dim)
{
    *res = vx[p];
}

typedef int bool;

"""

_header_c_double = _header_c_float.replace("float", "double")


class CompilationError(Exception):
    """
    Raised when a compilation error was detected.
    """
    pass


def compile_c_function(code_c, nbout, dtype=numpy.float32, add_header=True,
                       suffix="", additional_paths=None, tmpdir='.', fLOG=None):
    """
    Compiles a C function with :epkg:`cffi`.
    It takes one features vector.

    :param nbout: number of expected outputs
    :param code_c: code C
    :param dtype: numeric type to use
    :param add_header: add common function before compiling
    :param suffix: avoid avoid the same compiled module name
    :param additional_paths: additional paths to add to the module
    :param tmpdir: see below
    :param fLOG: logging function
    :return: compiled function

    The function assumes the first line is the signature.
    If you are using Windows with Visual Studio 2017, make sure
    you are using :epkg:`Python` 3.6.3+
    (see `Issue 30389 <https://bugs.python.org/issue30389>`_).
    Parameter *tmpdir* is used by function `compile
    <http://cffi.readthedocs.io/en/latest/cdef.html?
    highlight=compile#ffibuilder-compile-etc-compiling-out-of-line-modules>`_.
    """
    if sys.platform.startswith("win"):
        if "VS140COMNTOOLS" not in os.environ:  # pragma: no cover
            raise CompilationError(
                "Visual Studio is not installed.\n{0}".format(
                    "\n".join("{0}={1}".format(k, v) for k, v in sorted(os.environ.items()))))

    sig = code_c.split("\n")[0].strip() + ";"
    name = sig.split()[1]
    include_paths = []
    lib_paths = []
    if additional_paths is None:
        additional_paths = []

    # ~ if len(additional_paths) == 0 and sys.platform.startswith("win") and \
    # ~ 'VSSDK140Install' not in os.environ:  # last condition is for the installed VisualStudio.
    # ~ if fLOG:
    #~ fLOG("[compile_c_function] fix PATH for VS2017 on Windows")
    # ~ # Update environment variables.
    # ~ adds = [r"C:\Program Files (x86)\Microsoft Visual Studio 14.0\VC\bin\amd64",
    # ~ r"C:\Program Files (x86)\Windows Kits\10\bin\10.0.15063.0\x64"]
    # ~ vcvars64 = os.path.join(adds[0], 'vcvars64.bat')
    #~ subprocess.run(vcvars64)

    # ~ # Add paths for VS2017.
    # ~ includes = [r'C:\Program Files (x86)\Windows Kits\10\Include\10.0.15063.0\shared',
    #~ r'C:\Program Files (x86)\Microsoft Visual Studio 14.0\VC\include',
    # ~ r'C:\Program Files (x86)\Microsoft Visual Studio\2017\Community\SDK\ScopeCppSDK\SDK\include\ucrt']
    # ~ libs = [r'C:\Program Files (x86)\Microsoft Visual Studio 14.0\VC\lib\amd64',
    #~ r'C:\Program Files (x86)\Windows Kits\10\Lib\10.0.15063.0\um\x64',
    # ~ r'C:\Program Files (x86)\Windows Kits\10\Lib\10.0.15063.0\ucrt\x64']
    # ~ opaths = os.environ['PATH'].split(';')
    # ~ for add in adds:
    # ~ if os.path.exists(add) and add not in opaths:
    #~ additional_paths.append(add)
    # ~ oinc = os.environ.get('INCLUDE', '').split(';')
    # ~ for inc in includes:
    # ~ if os.path.exists(inc) and inc not in oinc:
    #~ include_paths.append(inc)
    # ~ for lib in libs:
    # ~ if os.path.exists(lib):
    #~ lib_paths.append(lib)

    if additional_paths:
        if fLOG:  # pragma: no cover
            for p in additional_paths:
                fLOG(f"[compile_c_function] PATH += '{p}'")
        os.environ["PATH"] += ";" + ";".join(additional_paths)

    if lib_paths and sys.platform.startswith("win"):  # pragma: no cover
        libs = ['msvcrt.lib', 'oldnames.lib', 'kernel32.lib', 'vcruntime.lib',
                'ucrt.lib']
        libs = {k: False for k in libs}
        for lib in lib_paths:
            for name in list(libs):
                if libs[name]:
                    continue
                msv = os.path.join(lib, name)
                if os.path.exists(msv):
                    dst = os.getcwd()
                    msvd = os.path.join(dst, name)
                    if not os.path.exists(msvd):
                        shutil.copy(msv, dst)
                        if fLOG:
                            fLOG(f"[compile_c_function] copy '{msv}'")
                    libs[name] = True
        copied = len([k for k, v in libs.items() if v])
        if copied < len(libs):
            raise CompilationError('Unable to find those libraries ({0}<{1}) {2} in\n{3}'.format(
                copied, len(libs), ','.join(sorted(libs)), '\n'.join(lib_paths)))

    if include_paths:
        if fLOG:  # pragma: no cover
            for p in include_paths:
                fLOG(f"[compile_c_function] INCLUDE += '{p}'")
        if 'INCLUDE' in os.environ:  # pragma: no cover
            os.environ["INCLUDE"] += ";" + ";".join(include_paths)
        else:  # pragma: no cover
            os.environ["INCLUDE"] = ";".join(include_paths)

    is_float = dtype == numpy.float32
    header = _header_c_float if is_float else _header_c_double
    code = code_c if not add_header else (header + code_c)

    from cffi import FFI
    ffibuilder = FFI()
    try:
        ffibuilder.cdef(sig)
    except Exception as e:  # pragma: no cover
        raise CompilationError(
            f"Signature is wrong\n{sig}\ndue to\n{e}") from e
    ffibuilder.set_source("_" + name + suffix, code)
    try:
        ffibuilder.compile(verbose=False, tmpdir=tmpdir)
    except Exception as e:  # pragma: no cover
        raise CompilationError(
            f"Compilation failed \n{sig}\ndue to\n{e}") from e
    mod = __import__(f"_{name}{suffix}")
    fct = getattr(mod.lib, name)

    def wrapper(features, output, cast_type, dtype):
        "wrapper for a vector of features"
        if len(features.shape) != 1:
            raise TypeError(  # pragma: no cover
                f"Only one dimension for the features not {features.shape}.")
        if output is None:
            output = numpy.zeros((nbout,), dtype=dtype)
        else:
            if len(output.shape) != 1:
                raise TypeError(  # pragma: no cover
                    f"Only one dimension for the output not {output.shape}.")
            if output.shape[0] != nbout:
                raise TypeError(  # pragma: no cover
                    f"Dimension mismatch {output.shape} != {nbout} (expected).")
            if output.dtype != dtype:
                raise TypeError(  # pragma: no cover
                    f"Type mismatch {output.dtype} != {dtype} (expected).")
        ptr = features.__array_interface__['data'][0]
        cptr = mod.ffi.cast(cast_type, ptr)
        optr = output.__array_interface__['data'][0]
        cout = mod.ffi.cast(cast_type, optr)
        fct(cout, cptr)
        return output

    def wrapper_double(features, output=None):
        "wrapper for double"
        return wrapper(features, output, "double*", numpy.float64)

    def wrapper_float(features, output=None):
        "wrapper for float"
        return wrapper(  # pragma: no cover
            features, output, "float*", numpy.float32)

    return wrapper_float if is_float else wrapper_double



# -*- encoding: utf-8 -*-
# pylint: disable=W0611
"""
@file
@brief Imports runtime operators.
"""

from ._op import OpRun
from .op_abs import Abs
from .op_acos import Acos
from .op_acosh import Acosh
from .op_adagrad import Adagrad
from .op_adam import Adam
from .op_add import Add
from .op_and import And
from .op_argmax import ArgMax
from .op_argmin import ArgMin
from .op_array_feature_extractor import ArrayFeatureExtractor
from .op_asin import Asin
from .op_asinh import Asinh
from .op_atan import Atan
from .op_atanh import Atanh
from .op_average_pool import AveragePool
from .op_batch_normalization import BatchNormalization, BatchNormalization_14
from .op_binarizer import Binarizer
from .op_bitshift import BitShift
from .op_broadcast_gradient_args import BroadcastGradientArgs
from .op_cast import Cast, CastLike
from .op_cdist import CDist
from .op_ceil import Ceil
from .op_celu import Celu
from .op_clip import Clip_6, Clip_11, Clip
from .op_category_mapper import CategoryMapper
from .op_complex_abs import ComplexAbs
from .op_compress import Compress
from .op_concat import Concat
from .op_concat_from_sequence import ConcatFromSequence
from .op_conv import Conv
from .op_conv_transpose import ConvTranspose
from .op_constant import Constant, Constant_12, Constant_11, Constant_9
from .op_constant_of_shape import ConstantOfShape
from .op_cos import Cos
from .op_cosh import Cosh
from .op_cum_sum import CumSum
from .op_debug import DEBUG
from .op_det import Det
from .op_depth_to_space import DepthToSpace, SpaceToDepth
from .op_dequantize_linear import DequantizeLinear
from .op_dict_vectorizer import DictVectorizer
from .op_div import Div
from .op_dropout import Dropout, Dropout_7, Dropout_12
from .op_einsum import Einsum
from .op_elu import Elu
from .op_equal import Equal
from .op_erf import Erf
from .op_exp import Exp
from .op_expand import Expand, Expand_13
from .op_expression import Expression
from .op_eyelike import EyeLike
from .op_feature_vectorizer import FeatureVectorizer
from .op_fft import FFT
from .op_fft2d import FFT2D
from .op_flatten import Flatten
from .op_fused_matmul import FusedMatMul
from .op_gather import Gather
from .op_gathernd import GatherND
from .op_gather_elements import GatherElements
from .op_gemm import Gemm
from .op_global_average_pool import GlobalAveragePool, GlobalMaxPool
from .op_greater import Greater, GreaterOrEqual
from .op_grid_sample import GridSample
from .op_gru import GRU
from .op_hardmax import Hardmax
from .op_hard_sigmoid import HardSigmoid
from .op_floor import Floor
from .op_identity import Identity
from .op_if import If
from .op_imputer import Imputer
from .op_inverse import Inverse
from .op_isinf import IsInf
from .op_isnan import IsNaN
from .op_label_encoder import LabelEncoder
from .op_leaky_relu import LeakyRelu
from .op_less import Less, LessOrEqual
from .op_linear_classifier import LinearClassifier
from .op_linear_regressor import LinearRegressor
from .op_log import Log
from .op_log_softmax import LogSoftmax
from .op_loop import Loop
from .op_lp_normalization import LpNormalization
from .op_lrn import LRN
from .op_lstm import LSTM
from .op_matmul import MatMul
from .op_max import Max
from .op_max_pool import MaxPool
from .op_mean import Mean
from .op_min import Min
from .op_mod import Mod
from .op_momentum import Momentum
from .op_mul import Mul
from .op_neg import Neg
from .op_negative_log_likelihood_loss import NegativeLogLikelihoodLoss
from .op_normalizer import Normalizer
from .op_non_max_suppression import NonMaxSuppression
from .op_non_zero import NonZero
from .op_not import Not
from .op_one_hot import OneHot
from .op_one_hot_encoder import OneHotEncoder
from .op_or import Or
from .op_pad import Pad
from .op_pow import Pow
from .op_prelu import PRelu
from .op_quantize_linear import QuantizeLinear, DynamicQuantizeLinear
from .op_qlinear_conv import QLinearConv
from .op_random import (
    Bernoulli, RandomNormal, RandomUniform,
    RandomUniformLike, RandomNormalLike)
from .op_range import Range
from .op_reciprocal import Reciprocal
from .op_reduce_log_sum import ReduceLogSum
from .op_reduce_log_sum_exp import ReduceLogSumExp
from .op_reduce_l1 import ReduceL1
from .op_reduce_l2 import ReduceL2
from .op_reduce_min import ReduceMin
from .op_reduce_max import ReduceMax
from .op_reduce_mean import ReduceMean
from .op_reduce_prod import ReduceProd
from .op_reduce_sum import (
    ReduceSum_1, ReduceSum_11, ReduceSum_13, ReduceSum)
from .op_reduce_sum_square import ReduceSumSquare
from .op_relu import Relu, ThresholdedRelu
from .op_reshape import Reshape, Reshape_5, Reshape_13, Reshape_14
from .op_resize import Resize
from .op_rfft import RFFT
from .op_roi_align import RoiAlign
from .op_round import Round
from .op_rnn import RNN
from .op_scaler import Scaler
from .op_scan import Scan
from .op_scatter_elements import ScatterElements
from .op_scatternd import ScatterND
from .op_softmax_cross_entropy_loss import SoftmaxCrossEntropyLoss
from .op_selu import Selu
from .op_sequence_at import SequenceAt
from .op_sequence_construct import SequenceConstruct
from .op_sequence_empty import SequenceEmpty
from .op_sequence_insert import SequenceInsert
from .op_shape import Shape
from .op_shrink import Shrink
from .op_sigmoid import Sigmoid
from .op_sign import Sign
from .op_sin import Sin
from .op_sinh import Sinh
from .op_size import Size
from .op_slice import Slice, Slice_1, Slice_10
from .op_split import Split, Split_2, Split_11, Split_13
from .op_softmax import Softmax, SoftmaxGrad, SoftmaxGrad_13
from .op_softplus import Softplus
from .op_softsign import Softsign
from .op_solve import Solve
from .op_sqrt import Sqrt
from .op_squeeze import Squeeze, Squeeze_1, Squeeze_11, Squeeze_13
from .op_string_normalizer import StringNormalizer
from .op_sub import Sub
from .op_sum import Sum
from .op_svm_classifier import SVMClassifier, SVMClassifierDouble
from .op_svm_regressor import SVMRegressor, SVMRegressorDouble
from .op_tan import Tan
from .op_tanh import Tanh
from .op_tfidfvectorizer import TfIdfVectorizer
from .op_tokenizer import Tokenizer
from .op_topk import TopK_10, TopK_11, TopK_1, TopK
from .op_transpose import Transpose
from .op_tree_ensemble_classifier import (
    TreeEnsembleClassifierDouble,
    TreeEnsembleClassifier_1, TreeEnsembleClassifier_3, TreeEnsembleClassifier)
from .op_tree_ensemble_regressor import (
    TreeEnsembleRegressorDouble,
    TreeEnsembleRegressor_1, TreeEnsembleRegressor_3, TreeEnsembleRegressor)
from .op_trilu import Trilu
from .op_unique import Unique
from .op_unsqueeze import Unsqueeze, Unsqueeze_1, Unsqueeze_11, Unsqueeze_13
from .op_where import Where
from .op_xor import Xor
from .op_yield_op import YieldOp
from .op_zipmap import ZipMap


from ..doc.doc_helper import get_rst_doc
_op_list = []
clo = locals().copy()
for name, cl in clo.items():
    if "_" in name:
        continue
    if name in {'cl', 'clo', 'name'}:
        continue  # pragma: no cover
    if not cl.__doc__ and issubclass(cl, OpRun):
        cl.__doc__ = get_rst_doc(cl.__name__)
        _op_list.append(cl)



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from ._op import OpRun
from .op_grid_sample_ import GridSampleFloat, GridSampleDouble  # pylint: disable=E0611


class GridSample(OpRun):

    atts = {'align_corners': 0,
            'mode': b'bilinear',
            'padding_mode': b'zeros'}

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       expected_attributes=GridSample.atts,
                       **options)
        self.rt32_ = None
        self.rt64_ = None
        self.rt32_ = GridSampleFloat()
        self.rt64_ = GridSampleDouble()
        self.rt32_.init(self.align_corners, self.mode, self.padding_mode)
        self.rt64_.init(self.align_corners, self.mode, self.padding_mode)

    def _run(self, X, grid, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        if X.dtype == numpy.float32:
            if self.rt32_ is None:
                self.rt32_ = GridSampleFloat()
                self.rt32_.init(self.align_corners,
                                self.mode, self.padding_mode)
            rt = self.rt32_
        elif X.dtype == numpy.float32:
            if self.rt64_ is None:
                self.rt64_ = GridSampleDouble()
                self.rt64_.init(self.align_corners,
                                self.mode, self.padding_mode)
            rt = self.rt64_
        else:
            raise TypeError(  # pragma: no cover
                f"Unsupported type {X.dtype!r} for GridSample.")

        res = rt.compute(X, grid)
        return (res, )



"""
@file
@brief Functions to manipulate ONNX file.
"""
from onnx import helper, AttributeProto


def find_node_name(model, name):
    """
    Finds a node by its name.
    :param model: onnx graph
    :param name: node name
    :return: node pointer
    """
    if not hasattr(model, "graph"):
        raise TypeError(  # pragma: no cover
            f"Parameter model is not an ONNX model but {type(model)}")
    for node in model.graph.node:
        if node.name == name:
            return node
    return None  # pragma: no cover


def find_node_input_name(node, name):
    """
    Finds a node input by its name.
    :param node: onnx node
    :param name: node name
    :return: input index
    """
    for i, inode in enumerate(node.input.node):
        if inode.name == name:
            return i
    return -1


def insert_node(model, op_type, node, input_index=0, new_name=None, **attrs):
    """
    Inserts a node before one node input.
    :param model: onnx graph
    :param op_type:
    :param node: node or node name
    :param input_index: input index or input name
    :param attrs: node attributes
    :return: updated graph
    """
    if isinstance(node, str):
        inode = find_node_name(model, node)
    else:
        inode = node
    if isinstance(input_index, str):
        input_index_ = find_node_input_name(node, input_index)
        if input_index_ == -1:
            raise RuntimeError(  # pragma: no cover
                "Unable to find input_index %r in node %r." % (
                    input_index, node.name))  # pylint: disable=E1120
        input_index = input_index_

    # guess a new name
    names = []
    for n in model.graph.node:
        names.extend(n.input)
        names.extend(n.output)
    names = set(names)
    if new_name is None:
        new_name = op_type.lower()
    root_name = new_name
    i = 0
    while new_name in names:
        new_name = "%s_%d" % (root_name, i)
        i += 1

    new_node = helper.make_node(
        op_type, [inode.input[input_index]], [new_name], **attrs)
    inode.input[input_index] = new_name
    keep_nodes = list(model.graph.node)
    keep_nodes.append(new_node)
    keep_nodes = ensure_topological_order(
        model.graph.input, model.graph.initializer, keep_nodes)

    graph = helper.make_graph(
        keep_nodes, model.graph.name, model.graph.input,
        model.graph.output, model.graph.initializer)
    onnx_model = helper.make_model(graph, functions=model.functions)
    onnx_model.ir_version = model.ir_version
    onnx_model.producer_name = model.producer_name
    onnx_model.producer_version = model.producer_version
    onnx_model.domain = model.domain
    onnx_model.model_version = model.model_version
    onnx_model.doc_string = model.doc_string
    if len(model.metadata_props) > 0:
        values = {p.key: p.value for p in model.metadata_props}
        helper.set_model_props(onnx_model, values)

    del onnx_model.opset_import[:]  # pylint: disable=E1101
    for oimp in model.opset_import:
        op_set = onnx_model.opset_import.add()  # pylint: disable=E1101
        op_set.domain = oimp.domain
        op_set.version = oimp.version

    if len(onnx_model.graph.input) != len(model.graph.input):  # pylint: disable=E1101
        raise RuntimeError(  # pragma: no cover
            "Input mismatch {} != {}".format(
                len(onnx_model.input), len(model.input)))  # pylint: disable=E1101
    return onnx_model


def ensure_topological_order(inputs, initializers, nodes):
    """
    Ensures and modifies the order of nodes to have
    a topological order (every node in the list
    can only be an input for a node later in this list).
    The function raises an exception if a cycle is detected.

    :param inputs: graph inputs:
    :param initializers: graph initializers
    :param nodes: graph nodes
    :return: list ordered nodes
    """
    order = {}
    for inp in inputs:
        name = inp.name
        order[name] = 0
    for inp in initializers:
        name = inp.name
        order[name] = 0
    n_iter = 0
    while n_iter < len(nodes) * 2:
        n_iter += 1
        missing_names = set()
        missing_ops = []
        for node in nodes:
            maxi = 0
            for name in node.input:
                if name in order:
                    maxi = max(maxi, order[name])
                else:
                    maxi = None
                    missing_names.add(name)
                    break
            if maxi is None:
                missing_ops.append(node)
                continue
            key = id(node)
            if key in order:
                continue
            maxi += 1
            order[key] = maxi
            maxi += 1
            for name in node.output:
                if name in order:
                    raise RuntimeError(  # pragma: no cover
                        "Unable to sort a node (cycle). An output was "
                        "already ordered %r (iteration=%r)." % (
                            name, n_iter))
                order[name] = maxi
        if len(missing_names) == 0:
            continue

    if len(missing_ops) > 0:  # pragma: no cover
        def nstr(name):
            if name in order:
                return "%s#%d" % (name, order[name])
            return name
        rows = ["%s(%s) -> [%s]" % (
            n.name or n.op_type,
            ', '.join(map(nstr, n.input)),
            ', '.join(n.output))
            for n in missing_ops]
        rows.insert(0, "")
        rows.append("--")
        rows.append("--all-nodes--")
        rows.append("--")
        rows.extend("%s(%s) -> [%s]" % (
            n.name or n.op_type,
            ', '.join(map(nstr, n.input)),
            ', '.join(n.output))
            for n in nodes)
        raise RuntimeError(
            "After %d iterations for %d nodes, still unable "
            "to sort names %r. The graph may be disconnected. "
            "List of operators: %s" % (
                n_iter, len(nodes), missing_names,
                "\n".join(rows)))

    # Update order
    topo = [(order[id(node)], str(id(node))) for node in nodes]
    topo.sort()
    map_nodes = {str(id(node)): node for node in nodes}
    return [map_nodes[_[1]] for _ in topo]


def enumerate_onnx_names(onx):
    """
    Enumerates all existing names in one ONNX graph
    (:epkg:`ModelProto`, :epkg:`FunctionProto`, :epkg:`GraphProto`).
    The function is recursive.

    :param onx: one onnx object
    :return: iterator on names
    """
    if hasattr(onx, 'graph'):
        for i in onx.graph.initializer:
            yield i.name
        for i in onx.graph.input:
            yield i.name
        for i in onx.graph.output:
            yield i.name
        nodes = onx.graph.node
    elif hasattr(onx, 'initializer'):
        for i in onx.initializer:
            yield i.name
        for i in onx.input:
            yield i.name
        for i in onx.output:
            yield i.name
        nodes = onx.node
    else:
        if hasattr(onx, 'input'):
            for i in onx.input:
                yield i
        if hasattr(onx, 'output'):
            for i in onx.output:
                yield i
        nodes = onx.node
    for node in nodes:
        for i in node.input:
            yield i
        for o in node.output:
            yield o
        for att in node.attribute:
            if (att.type == AttributeProto.GRAPH and  # pylint: disable=E0611,E1101
                    hasattr(att, 'g') and att.g is not None):
                for n in enumerate_onnx_names(att.g):
                    yield n


def enumerate_onnx_nodes(onx):
    """
    Enumerates all nodes in one ONNX graph
    (:epkg:`ModelProto`, :epkg:`FunctionProto`, :epkg:`GraphProto`).
    The function is recursive.

    :param onx: one onnx object
    :return: iterator on names
    """
    if isinstance(onx, list):
        nodes = onx
    elif hasattr(onx, 'graph'):
        nodes = onx.graph.node
    else:
        nodes = onx.node
    for node in nodes:
        yield node
        for att in node.attribute:
            if (att.type == AttributeProto.GRAPH and  # pylint: disable=E0611,E1101
                    hasattr(att, 'g') and att.g is not None):
                for n in enumerate_onnx_nodes(att.g):
                    yield n



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from ._op import OpRunBinaryComparison


class Equal(OpRunBinaryComparison):

    def __init__(self, onnx_node, desc=None, **options):
        OpRunBinaryComparison.__init__(
            self, onnx_node, desc=desc, **options)

    def _run(self, a, b, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        return (numpy.equal(a, b), )

    def to_python(self, inputs):
        return self._to_python_numpy(inputs, self.__class__.__name__.lower())



# pylint: disable=E1101
"""
@file
@brief Wrapper tokenizrs implemented in :epkg:`onnxruntime-extensions`.
"""
from io import BytesIO
import base64
import numpy
from scipy.sparse import csr_matrix
from sklearn.base import BaseEstimator, TransformerMixin
from onnx import helper, TensorProto, load
from onnx.defs import onnx_opset_version
try:
    from onnxruntime_extensions import get_library_path
except ImportError:  # pragma: no cover
    get_library_path = None
from mlprodict import __max_supported_opset__


class TokenizerTransformerBase(BaseEstimator, TransformerMixin):
    """
    Base class for @see cl SentencePieceTokenizerTransformer and
    @see cl GPT2TokenizerTransformer.
    """

    def __init__(self):
        BaseEstimator.__init__(self)
        TransformerMixin.__init__(self)
        from onnxruntime import InferenceSession, SessionOptions  # delayed
        self._InferenceSession = InferenceSession
        self._SessionOptions = SessionOptions

    def __getstate__(self):
        state = BaseEstimator.__getstate__(self)
        del state['sess_']
        del state['_InferenceSession']
        del state['_SessionOptions']
        state['onnx_'] = state['onnx_'].SerializeToString()
        return state

    def __setstate__(self, state):
        if get_library_path is None:
            raise ImportError(  # pragma: no cover
                "onnxruntime_extensions is not installed.")
        from onnxruntime import InferenceSession, SessionOptions  # delayed
        state['onnx_'] = load(BytesIO(state['onnx_']))
        BaseEstimator.__setstate__(self, state)
        self._InferenceSession = InferenceSession
        self._SessionOptions = SessionOptions
        so = SessionOptions()
        so.register_custom_ops_library(get_library_path())
        self.sess_ = InferenceSession(self.onnx_.SerializeToString(), so,
                                      providers=['CPUExecutionProvider'])
        return self


class SentencePieceTokenizerTransformer(TokenizerTransformerBase):
    """
    Wraps `SentencePieceTokenizer
    <https://github.com/microsoft/onnxruntime-extensions/blob/
    main/docs/custom_text_ops.md#sentencepiecetokenizer>`_
    into a :epkg:`scikit-learn` transformer.

    :param model: The sentencepiece model serialized proto as
        stored as a string
    :param nbest_size: tensor(int64) A scalar for sampling.
        `nbest_size = {0,1}`: no sampling is performed.
        (default) `nbest_size > 1`: samples from the nbest_size results.
        `nbest_size < 0`: assuming that nbest_size is infinite and
        samples from the all hypothesis (lattice) using
        forward-filtering-and-backward-sampling algorithm.
    :param alpha: tensor(float) A scalar for a smoothing parameter.
        Inverse temperature for probability rescaling.
    :param reverse: tensor(bool) Reverses the tokenized sequence.
    :param add_bos: tensor(bool) Add beginning of sentence token to the result.
    :param add_eos: tensor(bool) Add end of sentence token to the result
        When reverse=True beginning/end of sentence tokens are added
        after reversing
    :param opset: main opset to use

    Method *fit* produces the following attributes:

    * `onnx_`: onnx graph
    * `sess_`: :epkg:`InferenceSession` used to compute the inference
    """

    def __init__(self, model, nbest_size=1, alpha=0.5, reverse=False,
                 add_bos=False, add_eos=False, opset=None):
        TokenizerTransformerBase.__init__(self)
        if isinstance(model, bytes):
            self.model_b64 = model
        else:
            ints = model.tolist()
            b64 = base64.b64encode(ints)
            self.model_b64 = b64
        self.nbest_size = nbest_size
        self.alpha = alpha
        self.reverse = reverse
        self.add_bos = add_bos
        self.add_eos = add_eos
        self.opset = opset
        if get_library_path is None:
            raise ImportError(  # pragma: no cover
                "onnxruntime_extensions is not installed.")

    def fit(self, X, y=None, sample_weight=None):
        """
        The model is not trains this method is still needed to
        set the instance up and ready to transform.

        :param X: array of strings
        :param y: unused
        :param sample_weight: unused
        :return: self
        """
        self.onnx_ = self._create_model(
            self.model_b64, opset=self.opset)
        so = self._SessionOptions()
        so.register_custom_ops_library(get_library_path())
        self.sess_ = self._InferenceSession(self.onnx_.SerializeToString(), so)
        return self

    @staticmethod
    def _create_model(model_b64, domain='ai.onnx.contrib', opset=None):
        nodes = []
        mkv = helper.make_tensor_value_info
        nodes.append(helper.make_node(
            'SentencepieceTokenizer',
            inputs=['inputs', 'nbest_size', 'alpha', 'add_bos', 'add_eos',
                    'reverse'],
            outputs=['out0', 'out1'],
            model=model_b64,
            name='SentencepieceTokenizeOpName',
            domain='ai.onnx.contrib'))
        inputs = [
            mkv('inputs', TensorProto.STRING, [None]),
            mkv('nbest_size', TensorProto.INT64, [None]),
            mkv('alpha', TensorProto.FLOAT, [None]),
            mkv('add_bos', TensorProto.BOOL, [None]),
            mkv('add_eos', TensorProto.BOOL, [None]),
            mkv('reverse', TensorProto.BOOL, [None])]
        graph = helper.make_graph(
            nodes, 'SentencePieceTokenizerTransformer', inputs, [
                mkv('out0', TensorProto.INT32, [None]),
                mkv('out1', TensorProto.INT64, [None])])
        if opset is None:
            opset = min(__max_supported_opset__, onnx_opset_version())
        model = helper.make_model(graph, opset_imports=[
            helper.make_operatorsetid('', opset)])
        model.opset_import.extend([helper.make_operatorsetid(domain, 1)])
        return model

    def transform(self, X):
        """
        Applies the tokenizers on an array of strings.

        :param X: array to strings.
        :return: sparses matrix with n_features
        """
        out0, out1 = self.sess_.run(['out0', 'out1'],
                                    {'inputs': X, 'nbest_size': self.nbest_size, 'alpha': self.alpha,
                                     'add_bos': self.add_bos, 'add_eos': self.add_eos,
                                     'reverse': self.reverse})
        values = numpy.ones(out0.shape[0], dtype=numpy.float32)
        return csr_matrix((values, out0, out1))


class GPT2TokenizerTransformer(TokenizerTransformerBase):
    """
    Wraps `GPT2Tokenizer
    <https://github.com/microsoft/onnxruntime-extensions/blob/
    main/docs/custom_text_ops.md#gpt2tokenizer>`_
    into a :epkg:`scikit-learn` transformer.

    :param vocab: The content of the vocabulary file,
        its format is same with hugging face.
    :param merges: The content of the merges file,
        its format is same with hugging face.
    :param padding_length: When the input is a set of query,
        the tokenized result is ragged tensor, so we need to pad
        the tensor to tidy tensor and the *padding_length* indicates
        the strategy of the padding.
        When the *padding_length* equals -1, we will pad the tensor
        to length of longest row.
        When the *padding_length* is more than 0, we will pad the tensor
        to the number of padding_length.
    :param opset: main opset to use

    Method *fit* produces the following attributes:

    * `onnx_`: onnx graph
    * `sess_`: :epkg:`InferenceSession` used to compute the inference
    """

    def __init__(self, vocab, merges, padding_length=-1, opset=None):
        TokenizerTransformerBase.__init__(self)
        self.vocab = vocab
        self.merges = merges
        self.padding_length = padding_length
        self.opset = opset
        if get_library_path is None:
            raise ImportError(  # pragma: no cover
                "onnxruntime_extensions is not installed.")

    def fit(self, X, y=None, sample_weight=None):
        """
        The model is not trains this method is still needed to
        set the instance up and ready to transform.

        :param X: array of strings
        :param y: unused
        :param sample_weight: unused
        :return: self
        """
        self.onnx_ = self._create_model(
            self.vocab, self.merges, self.padding_length, opset=self.opset)
        so = self._SessionOptions()
        so.register_custom_ops_library(get_library_path())
        self.sess_ = self._InferenceSession(self.onnx_.SerializeToString(), so)
        return self

    @staticmethod
    def _create_model(vocab, merges, padding_length,
                      domain='ai.onnx.contrib', opset=None):
        nodes = []
        mkv = helper.make_tensor_value_info
        nodes.append(helper.make_node(
            'GPT2Tokenizer',
            inputs=['inputs'],
            outputs=['input_ids', 'attention_mask'],
            vocab=vocab, merges=merges,
            padding_length=padding_length,
            name='GPT2TokenizerName',
            domain='ai.onnx.contrib'))
        inputs = [mkv('inputs', TensorProto.STRING, [None])]
        graph = helper.make_graph(
            nodes, 'GPT2TokenizerTransformer', inputs, [
                mkv('input_ids', TensorProto.INT64, [None, None]),
                mkv('attention_mask', TensorProto.INT64, [None, None])])
        if opset is None:
            opset = min(__max_supported_opset__, onnx_opset_version())
        model = helper.make_model(
            graph, opset_imports=[helper.make_operatorsetid('', opset)])
        model.opset_import.extend([helper.make_operatorsetid(domain, 1)])
        return model

    def transform(self, X):
        """
        Applies the tokenizers on an array of strings.

        :param X: array to strings.
        :return: sparses matrix with n_features
        """
        input_ids, _ = self.sess_.run(
            ['input_ids', 'attention_mask'], {'inputs': X})
        idx = input_ids.ravel()
        values = numpy.ones(idx.shape[0], dtype=numpy.float32)
        rg = numpy.arange(input_ids.shape[0] + 1).astype(numpy.int64)
        rows = rg * input_ids.shape[1]
        return csr_matrix((values, idx, rows))



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from ._op import OpRunUnaryNum, OpRunBinaryNum
from ._new_ops import OperatorSchema


class Softmax(OpRunUnaryNum):

    atts = {'axis': 1}

    def __init__(self, onnx_node, desc=None, **options):
        OpRunUnaryNum.__init__(self, onnx_node, desc=desc,
                               expected_attributes=Softmax.atts,
                               **options)

    def _run(self, X, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        if self.inplaces.get(0, False) and X.flags['WRITEABLE']:
            return self._run_inplace(X)
        tmp = X - X.max(axis=self.axis, keepdims=1)
        Y = numpy.exp(tmp)
        Y /= Y.sum(axis=self.axis, keepdims=1)
        return (Y, )

    def _run_inplace(self, X):
        X -= X.max(axis=self.axis, keepdims=1)
        numpy.exp(X, out=X)
        X /= X.sum(axis=self.axis, keepdims=1)
        return (X, )

    def to_python(self, inputs):
        lines = ["tmp = {0} - {0}.max(axis=axis)[:, numpy.newaxis]".format(
            inputs[0]),
            "Y = numpy.exp(tmp)",
            "Y /= Y.sum(axis=axis)[:, numpy.newaxis]",
            "return Y"]
        return ("import numpy", "\n".join(lines))


class SoftmaxGrad_13(OpRunBinaryNum):
    """
    SoftmaxGrad computes :math:`dX = Y * ( dY - ReduceSum(Y * dY))`.
    ONNX does not have a dot product,
    which can be simulated as a pointwise-multiplication ("Mul"),
    followed by a "ReduceSum". Unfortunately, the treatment of "axis"
    is different in "SoftmaxGrad" and "ReduceSum".
    If axis=k for SoftmaxGrad, we need to specify [k, ..., n-1] as the axes of
    reduction for "ReduceSum", after accounting for negative-axis specification.
    An alternative solution would be to Flatten inputs to 2D and then reshape
    output back to original shape. Hopefully, many of these ops can be optimized
    away in the common-case of statically-known shapes.
    """

    atts = {'axis': 1}

    def __init__(self, onnx_node, desc=None, **options):
        OpRunBinaryNum.__init__(self, onnx_node, desc=desc,
                                expected_attributes=SoftmaxGrad_13.atts,
                                **options)

    def _find_custom_operator_schema(self, op_name):
        if op_name in ("SoftmaxGrad_13", "SoftmaxGrad"):
            return SoftmaxGradSchema()
        raise RuntimeError(  # pragma: no cover
            f"Unable to find a schema for operator '{op_name}'.")

    def _run(self, grad, prob, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        # softmax
        # tmp = X - X.max(axis=self.axis)[:, numpy.newaxis]
        # Y = numpy.exp(tmp)
        # Y /= Y.sum(axis=self.axis)[:, numpy.newaxis]
        # derivative
        pg = prob * grad
        if self.axis < 0:
            axis = len(pg.shape) + self.axis
        else:
            axis = self.axis
        axis = tuple(range(axis, len(pg.shape)))
        dg = grad - pg.sum(axis=axis, keepdims=1)
        return (prob * dg, )


class SoftmaxGradSchema(OperatorSchema):
    """
    Defines a schema for operators added in this package
    such as @see cl SoftmaxGrad_13.
    """

    def __init__(self):
        OperatorSchema.__init__(self, 'SoftmaxGrad')
        self.attributes = SoftmaxGrad_13.atts


SoftmaxGrad = SoftmaxGrad_13



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from ._op import OpRunUnaryNum


class Softplus(OpRunUnaryNum):

    def __init__(self, onnx_node, desc=None, **options):
        OpRunUnaryNum.__init__(self, onnx_node, desc=desc,
                               **options)

    def _run(self, X, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        if self.inplaces.get(0, False) and X.flags['WRITEABLE']:
            return self._run_inplace(X)
        tmp = numpy.exp(X)
        tmp += 1
        numpy.log(tmp, out=tmp)
        return (tmp, )

    def _run_inplace(self, X):
        numpy.exp(X, out=X)
        X += 1
        numpy.log(X, out=X)
        return (X, )

    def to_python(self, inputs):
        lines = [f"Y = numpy.exp({inputs[0]})",
                 "Y += 1",
                 "numpy.log(Y, out=Y)",
                 "return Y"]
        return ("import numpy", "\n".join(lines))



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from ._op import OpRun
from ._new_ops import OperatorSchema


class ComplexAbs(OpRun):

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc, **options)

    def _find_custom_operator_schema(self, op_name):
        if op_name == "ComplexAbs":
            return ComplexAbsSchema()
        raise RuntimeError(  # pragma: no cover
            f"Unable to find a schema for operator '{op_name}'.")

    def _run(self, x, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        y = numpy.absolute(x)
        if x.dtype == numpy.complex64:
            y = y.astype(numpy.float32)
        elif x.dtype == numpy.complex128:
            y = y.astype(numpy.float64)
        else:
            raise TypeError(  # pragma: no cover
                f"Unexpected input type for x: {x.dtype!r}.")
        return (y, )

    def to_python(self, inputs):
        return self._to_python_numpy(inputs, 'absolute')


class ComplexAbsSchema(OperatorSchema):
    """
    Defines a schema for operators added in this package
    such as @see cl ComplexAbs.
    """

    def __init__(self):
        OperatorSchema.__init__(self, 'ComplexAbs')
        self.attributes = {}



"""
@file
@brief Custom operators for complex numbers.
"""
from skl2onnx.algebra.onnx_operator import OnnxOperator


class OnnxComplexAbs_1(OnnxOperator):
    """
    Defines a custom operator for ComplexAbs.
    """

    since_version = 1
    expected_inputs = [('X', 'T'), ]
    expected_outputs = [('Y', 'U')]
    input_range = [1, 1]
    output_range = [1, 1]
    is_deprecated = False
    domain = 'mlprodict'
    operator_name = 'ComplexAbs'
    past_version = {}

    def __init__(self, X, op_version=None, **kwargs):
        """
        :param X: array or OnnxOperatorMixin
        :param op_version: opset version
        :param kwargs: additional parameter
        """
        OnnxOperator.__init__(
            self, X, op_version=op_version, **kwargs)


OnnxComplexAbs = OnnxComplexAbs_1



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
from ...onnx_tools.onnx2py_helper import guess_dtype
from ._op import OpRun


class If(OpRun):

    atts = {
        'then_branch': None,
        'else_branch': None,
    }

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       expected_attributes=If.atts,
                       **options)
        if not hasattr(self.then_branch, 'run'):
            raise RuntimeError(  # pragma: no cover
                "Parameter 'then_branch' must have a method 'run', "
                "type {}.".format(type(self.then_branch)))
        if not hasattr(self.else_branch, 'run'):
            raise RuntimeError(  # pragma: no cover
                "Parameter 'else_branch' must have a method 'run', "
                "type {}.".format(type(self.else_branch)))

        self._run_meth_then = (self.then_branch.run_in_scan
                               if hasattr(self.then_branch, 'run_in_scan')
                               else self.then_branch.run)
        self._run_meth_else = (self.else_branch.run_in_scan
                               if hasattr(self.else_branch, 'run_in_scan')
                               else self.else_branch.run)
        self.additional_inputs = list(
            set(self.then_branch.static_inputs) |
            set(self.else_branch.static_inputs))

    def need_context(self):
        """
        Tells the runtime if this node needs the context
        (all the results produced so far) as it may silently access
        one of them (operator Loop).
        The default answer is `False`.
        """
        return True

    def _run(self, cond, named_inputs=None, context=None,  # pylint: disable=W0221
             attributes=None, verbose=0, fLOG=None):
        if cond is None:
            raise RuntimeError(  # pragma: no cover
                "cond cannot be None")
        if named_inputs is None:
            named_inputs = {}
        if len(self.then_branch.input_names) > 0:
            if len(context) == 0:
                raise RuntimeError(  # pragma: no cover
                    "named_inputs is empty but the graph needs {}, "
                    "sub-graphs for node If must not have any inputs.".format(
                        self.then_branch.input_names))
            for k in self.then_branch.input_names:
                if k not in context:
                    raise RuntimeError(  # pragma: no cover
                        "Unable to find named input '{}' in\n{}.".format(
                            k, "\n".join(sorted(context))))
        if len(self.else_branch.input_names) > 0:
            if len(context) == 0:
                raise RuntimeError(  # pragma: no cover
                    "context is empty but the graph needs {}.".format(
                        self.then_branch.input_names))
            for k in self.else_branch.input_names:
                if k not in context:
                    raise RuntimeError(  # pragma: no cover
                        "Unable to find named input '{}' in\n{}.".format(
                            k, "\n".join(sorted(context))))

        # then_local_inputs = set(self.local_inputs(self.then_branch.obj.graph))
        # else_local_inputs = set(self.local_inputs(self.else_branch.obj.graph))
        # self.additional_inputs = list(
        #     set(self.additional_inputs).union(then_local_inputs.union(else_local_inputs)))
        # for n in self.additional_inputs:
        #     self.then_branch.global_index(n)
        #     self.else_branch.global_index(n)

        if len(cond.shape) > 0:
            if all(cond):
                if verbose > 0 and fLOG is not None:
                    fLOG(  # pragma: no cover
                        f'  -- then> {list(context)!r}')
                outputs = self._run_meth_then(named_inputs, context=context,
                                              attributes=attributes,
                                              verbose=verbose, fLOG=fLOG)
                if verbose > 0 and fLOG is not None:
                    fLOG('  -- then<')
                final = tuple([outputs[name]
                              for name in self.then_branch.output_names])
                branch = 'then'
            else:
                if verbose > 0 and fLOG is not None:
                    fLOG(  # pragma: no cover
                        f'  -- else> {list(context)!r}')
                outputs = self._run_meth_else(named_inputs, context=context,
                                              attributes=attributes,
                                              verbose=verbose, fLOG=fLOG)
                if verbose > 0 and fLOG is not None:
                    fLOG('  -- else<')  # pragma: no cover
                final = tuple([outputs[name]
                              for name in self.else_branch.output_names])
                branch = 'else'
        elif cond:
            if verbose > 0 and fLOG is not None:
                fLOG(  # pragma: no cover
                    f'  -- then> {list(context)!r}')
            outputs = self._run_meth_then(named_inputs, context=context,
                                          attributes=attributes,
                                          verbose=verbose, fLOG=fLOG)
            if verbose > 0 and fLOG is not None:
                fLOG('  -- then<')  # pragma: no cover
            final = tuple([outputs[name]
                          for name in self.then_branch.output_names])
            branch = 'then'
        else:
            if verbose > 0 and fLOG is not None:
                fLOG(  # pragma: no cover
                    f'  -- else> {list(context)!r}')
            outputs = self._run_meth_else(named_inputs, context=context,
                                          attributes=attributes,
                                          verbose=verbose, fLOG=fLOG)
            if verbose > 0 and fLOG is not None:
                fLOG('  -- else<')  # pragma: no cover
            final = tuple([outputs[name]
                          for name in self.else_branch.output_names])
            branch = 'else'

        if len(final) == 0:
            raise RuntimeError(  # pragma: no cover
                f"Operator If ({self.onnx_node.name!r}) does not have any output.")
        for i, f in enumerate(final):
            if f is None:
                ni = named_inputs if named_inputs else []  # pragma: no cover
                br = self.then_branch if branch == 'then' else self.else_branch
                names = br.output_names
                inits = [i.name for i in br.obj.graph.initializer]
                raise RuntimeError(  # pragma: no cover
                    "Output %d (branch=%r, name=%r) is None, available inputs=%r, "
                    "initializers=%r." % (
                        i, branch, names[i], list(sorted(ni)), inits))
        return final

    def _pick_type(self, res, name):
        if name in res:
            return res[name]
        out = {o.name: o for o in self.then_branch.obj.graph.output}
        if name not in out:
            raise ValueError(  # pragma: no cover
                "Unable to find name=%r in %r or %r." % (
                    name, list(sorted(res)), list(sorted(out))))
        dt = out[name].type.tensor_type.elem_type
        return guess_dtype(dt)



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import math
import numpy
from ._op import OpRun


class LRN(OpRun):

    atts = {
        'alpha': 9.999999747378752e-05,
        'beta': 0.75,
        'bias': 1.,
        'size': 3,
    }

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       expected_attributes=LRN.atts,
                       **options)

    def _run(self, x, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        if len(x.shape) != 4:
            raise RuntimeError(  # pragma: no cover
                f"LRN only applies on 4D tensors but shape is {x.shape!r}.")
        square_sum = numpy.zeros(x.shape).astype(x.dtype)
        for ind in numpy.ndindex(x.shape):
            n, c, h, w = ind
            begin = max(0, c - int(math.floor((self.size - 1) / 2)))
            end = min(5, c + int(math.ceil((self.size - 1) / 2)) + 1)
            square_sum[n, c, h, w] = numpy.sum(x[n, begin:end, h, w] ** 2)
        y = x / ((self.bias + (self.alpha / self.size) * square_sum) ** self.beta)
        return (y.astype(x.dtype), )



"""
@file
@brief Runtime operator.
"""
import numpy
from ._op import OpRun


def _gather_nd_impl(data, indices, batch_dims):
    """
    Modified version of `softmaxcrossentropy.py
    <https://github.com/onnx/onnx/blob/main/onnx/backend/
    test/case/node/gathernd.py>`_.
    """
    # Note the data rank - will be reused multiple times later
    data_rank = len(data.shape)

    # The list of data/indice shape of batch_dims.
    batch_dims_shape = []

    # The number of elements in the batch_dims for data/indice array.
    batch_dims_size = 1

    # Check the shape of indice and data are identicial for batch dims.
    for i in range(batch_dims):
        batch_dims_shape.append(indices.shape[i])
        batch_dims_size *= indices.shape[i]

    # Compute output of the op as below.
    # Compute shape of output array.
    output_shape = (
        batch_dims_shape + list(indices.shape)[batch_dims:-1]
        if (indices.shape[-1] == data_rank - batch_dims)
        else batch_dims_shape + list(indices.shape)[batch_dims:-1] +
        list(data.shape)[batch_dims + indices.shape[-1]:])

    # Placeholder for output data.
    output_data_buffer = []

    # Flatten 'indices' to 2D array.
    reshaped_indices = indices.reshape(batch_dims_size, -1, indices.shape[-1])

    # Flatten 'data' to array of shape
    # (batch_dim_size, data.shape[batch_dimes:]).
    reshaped_data = data.reshape((batch_dims_size, ) + data.shape[batch_dims:])

    # Gather each scalar value from 'data'.
    for batch_dim in range(reshaped_indices.shape[0]):
        for outer_dim in range(reshaped_indices.shape[1]):
            gather_index = tuple(reshaped_indices[batch_dim][outer_dim])
            output_data_buffer.append(
                reshaped_data[(batch_dim,) + gather_index])
    return (numpy.asarray(output_data_buffer,
                          dtype=data.dtype).reshape(output_shape), )


class GatherND(OpRun):
    """
    Python runtime for function *SoftmaxCrossEntropyLoss*.
    """

    atts = {'batch_dims': 0}

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       expected_attributes=GatherND.atts,
                       **options)

    def _run(self, data, indices, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        return _gather_nd_impl(data, indices, self.batch_dims)  # pylint: disable=E1101



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
from ._op import OpRunUnary


class Scaler(OpRunUnary):

    atts = {'offset': None, 'scale': None}

    def __init__(self, onnx_node, desc=None, **options):
        OpRunUnary.__init__(self, onnx_node, desc=desc,
                            expected_attributes=Scaler.atts,
                            **options)

    def _run(self, x, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        return self._run_no_checks_(x, verbose=verbose, fLOG=fLOG)

    def _run_no_checks_(self, x, verbose=0, fLOG=None):  # pylint: disable=W0221
        if self.inplaces.get(0, False) and x.flags['WRITEABLE']:
            return self._run_inplace(x)
        dx = x - self.offset
        return (dx * self.scale, )

    def _run_inplace(self, x):
        x -= self.offset
        x *= self.scale
        return (x, )



# -*- encoding: utf-8 -*-
"""
@file
@brief Python runtime for ONNX and others tools to help
converting investigate issues with ONNX models.
"""

__version__ = "0.8.1858"
__author__ = "Xavier Dupré"
__max_supported_opset__ = 15  # Converters are tested up to this version.
__max_supported_opsets__ = {
    '': __max_supported_opset__,
    'ai.onnx.ml': 2}
# Converters are tested up to this version.
__max_supported_opset_experimental__ = 16
__max_supported_opsets_experimental__ = {
    '': __max_supported_opset_experimental__,
    'ai.onnx.ml': 3}


def get_ir_version(opv):
    """
    Returns the corresponding `IR_VERSION` based on the selected opset.
    See :epkg:`ONNX Version`.

    :param opv: opset
    :return: runtime version
    """
    if isinstance(opv, dict):
        opv = opv['']
    if opv is None or opv >= 15:
        return 8
    if opv >= 12:
        return 7
    if opv >= 11:  # pragma no cover
        return 6
    if opv >= 10:  # pragma no cover
        return 5
    if opv >= 9:  # pragma no cover
        return 4
    if opv >= 8:  # pragma no cover
        return 4
    return 3  # pragma no cover


def check(log=False):
    """
    Checks the library is working.
    It raises an exception.
    If you want to disable the logs:

    @param      log     if True, display information, otherwise
    @return             0 or exception
    """
    return True


def load_ipython_extension(ip):  # pragma: no cover
    """
    To allow the call ``%load_ext mlprodict``

    @param      ip      from ``get_ipython()``
    """
    from .nb_helper import register_onnx_magics as freg
    freg(ip)



# pylint: disable=E0602
"""
@file
@brief Xop API to build onnx graphs. Inspired from :epkg:`sklearn-onnx`.

.. versionadded:: 0.9
"""
import numpy
from .xop import loadop


def OnnxReduceSumApi11(*x, axes=None, keepdims=1, op_version=None,
                       output_names=None):
    """
    Adds operator ReduceSum with opset>=13 following API from opset 12.
    """
    if op_version is None:
        raise RuntimeError(  # pragma: no cover
            "op_version must be specified.")
    if op_version is None or op_version >= 13:
        OnnxReduceSum = loadop('ReduceSum')
        if axes is None:
            return OnnxReduceSum(
                *x, keepdims=keepdims, op_version=op_version,
                output_names=output_names)
        return OnnxReduceSum(
            *x, numpy.array(axes, dtype=numpy.int64),
            keepdims=keepdims, op_version=op_version,
            output_names=output_names)
    if op_version >= 11:
        OnnxReduceSum_11 = loadop('ReduceSum_11')
        if axes is None:
            return OnnxReduceSum_11(
                *x, keepdims=keepdims,
                op_version=op_version, output_names=output_names)
        return OnnxReduceSum_11(
            *x, axes=axes, keepdims=keepdims,
            op_version=op_version, output_names=output_names)
    OnnxReduceSum_1 = loadop('ReduceSum_1')
    if axes is None:
        return OnnxReduceSum_1(*x, keepdims=keepdims,
                               op_version=op_version,
                               output_names=output_names)
    return OnnxReduceSum_1(*x, axes=axes, keepdims=keepdims,
                           op_version=op_version, output_names=output_names)


def OnnxSplitApi11(*x, axis=0, split=None, op_version=None,
                   output_names=None):
    """
    Adds operator Split with opset>=13 following API from opset 11.
    """
    if op_version is None:
        raise RuntimeError(  # pragma: no cover
            "op_version must be specified.")
    if op_version is None or op_version >= 13:
        OnnxSplit = loadop('Split')
        if split is None:
            return OnnxSplit(
                *x, axis=axis, op_version=op_version,
                output_names=output_names)
        return OnnxSplit(
            *x, numpy.array(split, dtype=numpy.int64), axis=axis,
            op_version=op_version, output_names=output_names)
    if op_version >= 11:
        OnnxSplit_11 = loadop('Split_11')
        if split is None:
            return OnnxSplit_11(
                *x, axis=axis, op_version=op_version,
                output_names=output_names)
        return OnnxSplit_11(
            *x, split=split, axis=axis, op_version=op_version,
            output_names=output_names)
    OnnxSplit_2 = loadop('Split_2')
    if split is None:
        return OnnxSplit_2(
            *x, axis=axis, op_version=op_version, output_names=output_names)
    return OnnxSplit_2(*x, split=split, axis=axis,
                       op_version=op_version, output_names=output_names)


def OnnxSqueezeApi11(*x, axes=None, op_version=None,
                     output_names=None):
    """
    Adds operator Squeeze with opset>=13 following API from opset 11.
    """
    if op_version is None:
        raise RuntimeError(  # pragma: no cover
            "op_version must be specified.")
    if op_version is None or op_version >= 13:
        OnnxSqueeze = loadop('Squeeze')
        return OnnxSqueeze(
            *x, numpy.array(axes, dtype=numpy.int64),
            op_version=op_version, output_names=output_names)
    if op_version >= 11:
        OnnxSqueeze_11 = loadop('Squeeze_11')
        return OnnxSqueeze_11(
            *x, axes=axes, op_version=op_version,
            output_names=output_names)
    OnnxSqueeze_1 = loadop('Squeeze_1')
    return OnnxSqueeze_1(*x, axes=axes,
                         op_version=op_version, output_names=output_names)


def OnnxUnsqueezeApi11(*x, axes=None, op_version=None,
                       output_names=None):
    """
    Adds operator Unsqueeze with opset>=13 following API from opset 11.
    """
    if op_version is None:
        raise RuntimeError(  # pragma: no cover
            "op_version must be specified.")
    if op_version is None or op_version >= 13:
        OnnxUnsqueeze = loadop('Unsqueeze')
        return OnnxUnsqueeze(
            *x, numpy.array(axes, dtype=numpy.int64),
            op_version=op_version, output_names=output_names)
    if op_version >= 11:
        OnnxUnsqueeze_11 = loadop('Unsqueeze_11')
        return OnnxUnsqueeze_11(
            *x, axes=axes, op_version=op_version,
            output_names=output_names)
    OnnxUnsqueeze_1 = loadop('Unsqueeze_1')
    return OnnxUnsqueeze_1(*x, axes=axes,
                           op_version=op_version, output_names=output_names)


def OnnxReduceL2_typed(dtype, x, axes=None, keepdims=1, op_version=None,
                       output_names=None):
    """
    Adds operator ReduceL2 for float or double.
    """
    OnnxMul, OnnxSqrt = loadop('Mul', 'Sqrt')
    if dtype == numpy.float32:
        OnnxReduceL2 = loadop('ReduceL2')
        return OnnxReduceL2(
            x, axes=axes, keepdims=keepdims,
            op_version=op_version, output_names=output_names)
    x2 = OnnxMul(x, x, op_version=op_version)
    red = OnnxReduceSumApi11(
        x2, axes=[1], keepdims=1, op_version=op_version)
    return OnnxSqrt(
        red, op_version=op_version, output_names=output_names)


def OnnxReshapeApi13(*x, allowzero=0, op_version=None,
                     output_names=None):
    """
    Adds operator Reshape with opset>=14 following API from opset 13.
    """
    if op_version is None:
        raise RuntimeError(  # pragma: no cover
            "op_version must be specified.")
    if op_version is None or op_version >= 14:
        OnnxReshape = loadop('Reshape')
        return OnnxReshape(
            *x, allowzero=allowzero,
            op_version=op_version, output_names=output_names)
    if op_version >= 13:
        OnnxReshape_13 = loadop('Reshape_13')
        return OnnxReshape_13(
            *x, op_version=op_version, output_names=output_names)
    OnnxReshape_5 = loadop('Reshape_5')
    return OnnxReshape_5(
        *x, op_version=op_version, output_names=output_names)



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from ._op import OpRunUnaryNum


class Tanh(OpRunUnaryNum):

    def __init__(self, onnx_node, desc=None, **options):
        OpRunUnaryNum.__init__(self, onnx_node, desc=desc,
                               **options)

    def _run(self, x, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        if self.inplaces.get(0, False) and x.flags['WRITEABLE']:
            return self._run_inplace(x)
        return (numpy.tanh(x), )

    def _run_inplace(self, x):
        return (numpy.tanh(x, out=x), )

    def to_python(self, inputs):
        return self._to_python_numpy(inputs, self.__class__.__name__.lower())



"""
@file
@brief Errors and exceptions for @see cl OnnxShapeInference.
"""


class ShapeInferenceException(RuntimeError):
    """
    Raised when shape inference fails.
    """
    pass


class ShapeInferenceMissing(RuntimeError):
    """
    Raised when an operator is missing.
    """
    pass


class NotImplementedShapeInferenceError(NotImplementedError):
    """
    Shape Inference can be implemented but is currently not.
    """
    pass


class ShapeInferenceDimensionError(RuntimeError):
    """
    Raised when the shape cannot continue
    due to unknown dimension.
    """
    pass



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from ._op import OpRunUnaryNum


class Shrink(OpRunUnaryNum):

    atts = {'bias': 0, 'lambd': 0.5}

    def __init__(self, onnx_node, desc=None, **options):
        OpRunUnaryNum.__init__(self, onnx_node, desc=desc,
                               expected_attributes=Shrink.atts,
                               **options)

    def _run(self, x, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        return (numpy.where(x < -self.lambd, x + self.bias,
                            numpy.where(x > self.lambd, x - self.bias, 0)), )

    def to_python(self, inputs):
        return (
            "import numpy",
            ("return numpy.where({0} < -lambd, {0} + bias, "
             "numpy.where({0} > lambd, {0} - bias, 0))").format(inputs[0]))



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from ._op import OpRunUnaryNum
from ._op_numpy_helper import numpy_dot_inplace


class LinearRegressor(OpRunUnaryNum):

    atts = {'coefficients': None, 'intercepts': None,
            'targets': 1, 'post_transform': b'NONE'}

    def __init__(self, onnx_node, desc=None, **options):
        OpRunUnaryNum.__init__(self, onnx_node, desc=desc,
                               expected_attributes=LinearRegressor.atts,
                               **options)
        if not isinstance(self.coefficients, numpy.ndarray):
            raise TypeError(  # pragma: no cover
                f"coefficient must be an array not {type(self.coefficients)}.")
        n = self.coefficients.shape[0] // self.targets
        self.coefficients = self.coefficients.reshape(self.targets, n).T

    def _run(self, x, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        score = numpy_dot_inplace(self.inplaces, x, self.coefficients)
        if self.intercepts is not None:
            score += self.intercepts
        if self.post_transform == b'NONE':
            pass
        else:
            raise NotImplementedError(  # pragma: no cover
                f"Unknown post_transform: '{self.post_transform}'.")
        return (score, )



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from ._op import OpRun, DefaultNone


class Compress(OpRun):

    atts = {'axis': DefaultNone}

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       expected_attributes=Compress.atts,
                       **options)

    def _run(self, x, condition, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        if self.inplaces.get(0, False) and x.flags['WRITEABLE']:
            return (numpy.compress(condition, x, axis=self.axis, out=x), )
        return (numpy.compress(condition, x, axis=self.axis), )

    def to_python(self, inputs):
        if self.axis is None:
            return "import numpy\nreturn numpy.compress(%s, %s)" % tuple(inputs)
        return "import numpy\nreturn numpy.compress(%s, %s, axis=%d)" % (
            tuple(inputs) + (self.axis, ))



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from ._op import OpRunUnaryNum


class Selu(OpRunUnaryNum):

    atts = {'alpha': 1.67326319217681884765625,
            'gamma': 1.05070102214813232421875}

    def __init__(self, onnx_node, desc=None, **options):
        OpRunUnaryNum.__init__(self, onnx_node, desc=desc,
                               expected_attributes=Selu.atts,
                               **options)

    def _run(self, x, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        return (numpy.where(
            x > 0, x,
            numpy.exp(x) * self.alpha - self.alpha) * self.gamma, )

    def to_python(self, inputs):
        return (
            "import numpy",
            ("return numpy.where({0} > 0, {0}, "
             "numpy.exp({0}) * {2} - {2}) * {1}").format(
                inputs[0], self.gamma, self.alpha))



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from ._op import OpRunReduceNumpy


class ReduceMax(OpRunReduceNumpy):

    atts = {'axes': [], 'keepdims': 1}

    def __init__(self, onnx_node, desc=None, **options):
        OpRunReduceNumpy.__init__(self, onnx_node, desc=desc,
                                  expected_attributes=ReduceMax.atts,
                                  **options)

    def _run(self, data, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        axes = tuple(self.axes) if self.axes else None
        return (numpy.maximum.reduce(data, axis=axes,  # pylint: disable=E1123
                                     keepdims=self.keepdims == 1), )



"""
@file
@brief Xop API. Importing this file takes time. It should be avoided.

.. versionadded:: 0.9
"""
import sys
from .xop import _dynamic_class_creation


def _update_module():
    """
    Dynamically updates the module with operators defined by *ONNX*.
    """
    res = _dynamic_class_creation(include_past=True)
    this = sys.modules[__name__]
    unique = set()
    for cl in res:
        setattr(this, cl.__name__, cl)
        unique.add((cl.domain, cl.operator_name))
    res = _dynamic_class_creation(list(unique))
    for cl in res:
        setattr(this, cl.__name__, cl)


_update_module()



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from ._op import OpRun


class Trilu(OpRun):

    atts = {'upper': 1}

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       expected_attributes=Trilu.atts,
                       **options)
        if self.upper not in (0, 1):
            raise ValueError(f"upper must be 0 or 1 not {self.upper!r}.")

    def _run(self, *inputs, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        x = inputs[0]
        k = 0 if len(inputs) == 1 else int(inputs[1])
        if self.upper:
            return (numpy.triu(x, k), )
        return (numpy.tril(x, k), )

    def to_python(self, inputs):
        name = "triu" if self.upper else "tril"
        return (
            "import numpy",
            "return numpy.%s(%s, int(%s))" % (
                name, inputs[0], 0 if len(inputs) == 1 else inputs[1]))



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from ._op import OpRun


class Einsum(OpRun):

    atts = {'equation': ''}
    python_inputs = ['*inputs']

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       expected_attributes=Einsum.atts,
                       **options)
        if not isinstance(self.equation, (str, bytes)):
            raise TypeError(  # pragma: no cover
                f"equation must be string but is {type(self.equation)!r}.")
        self.equation = self.equation.strip()
        if len(self.equation) == 0:
            raise TypeError("equation is empty.")  # pragma: no cover

    def _run(self, *args, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        try:
            return (numpy.einsum(self.equation, *args, optimize=True), )
        except TypeError:
            return (numpy.einsum(self.equation, *args), )

    def to_python(self, inputs):
        return ("import numpy",
                "return numpy.einsum(equation, *inputs)")



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from ._op import OpRun


class FeatureVectorizer(OpRun):
    """
    Very similar to @see cl Concat.
    """

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       **options)
        self.axis = 1

    def _preprocess(self, a):
        if self.axis >= len(a.shape):
            new_shape = a.shape + (1, ) * (self.axis + 1 - len(a.shape))
            return a.reshape(new_shape)
        return a

    def _run(self, *args, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        args = [self._preprocess(a) for a in args]
        res = numpy.concatenate(args, self.axis)
        return (res, )



"""
@file
@brief Command line to check einsum scenarios.
"""


def onnx_code(filename, format="onnx", output=None, verbose=0, name=None,
              opset=None, fLOG=print):
    """
    Exports an ONNX graph into a python code creating
    the same graph.

    :param filename: onnx file
    :param format: format to export too (`onnx`, `tf2onnx`, `numpy`)
    :param output: output file to produce or None to print it on stdout
    :param verbose: verbosity level
    :param name: rewrite the graph name
    :param opset: overwrite the opset (may not works depending on the format)
    :param fLOG: logging function

    .. cmdref::
        :title: Exports an ONNX graph into a python code creating the same graph.
        :cmd: -m mlprodict onnx_code --help
        :lid: l-cmd-onnx_code

        The command converts an ONNX graph into a python code generating
        the same graph. The python code may use onnx syntax, numpy syntax
        or tf2onnx syntax.

        Example::

            python -m mlprodict onnx_code --filename="something.onnx" --format=onnx
    """
    from ..onnx_tools.onnx_export import (  # pylint: disable=E0402
        export2onnx, export2tf2onnx, export2numpy)

    if name == '':
        name = None
    if opset == '':
        opset = None
    try:
        v = int(opset)
        opset = v
    except (ValueError, TypeError):
        opset = None

    if format == 'onnx':
        code = export2onnx(filename, verbose=verbose, name=name, opset=opset)
    elif format == 'tf2onnx':
        code = export2tf2onnx(filename, verbose=verbose,
                              name=name, opset=opset)
    elif format == 'numpy':
        code = export2numpy(filename, verbose=verbose,
                            name=name, opset=opset)
    else:
        raise ValueError(  # pragma: no cover
            f"Unknown format {format!r}.")

    if output not in ('', None):
        with open(output, "w", encoding="utf-8") as f:
            f.write(code)
    else:
        fLOG(code)  # pragma: no cover


def dynamic_doc(verbose=0, fLOG=print):
    """
    Generates the documentation for ONNX operators.

    :param verbose: displays the list of operator
    :param fLOG: logging function
    """
    from ..npy.xop import _dynamic_class_creation
    _dynamic_class_creation(cache=True, verbose=verbose, fLOG=fLOG)


def plot_onnx(filename, format="onnx", verbose=0, output=None, fLOG=print):
    """
    Plots an ONNX graph on the standard output.

    :param filename: onnx file
    :param format: format to export too (`simple`, `tree`, `dot`,
        `io`, `mat`, `raw`)
    :param output: output file to produce or None to print it on stdout
    :param verbose: verbosity level
    :param fLOG: logging function

    .. cmdref::
        :title: Plots an ONNX graph as text
        :cmd: -m mlprodict plot_onnx --help
        :lid: l-cmd-plot_onnx

        The command shows the ONNX graphs as a text on the standard output.

        Example::

            python -m mlprodict plot_onnx --filename="something.onnx" --format=simple
    """
    if isinstance(filename, str):
        from onnx import load
        content = load(filename)
    else:
        content = filename
    if format == 'dot':
        from ..onnxrt import OnnxInference
        code = OnnxInference(filename).to_dot()
    elif format == 'simple':
        from mlprodict.plotting.text_plot import onnx_simple_text_plot
        code = onnx_simple_text_plot(content)
    elif format == 'io':
        from mlprodict.plotting.text_plot import onnx_text_plot_io
        code = onnx_text_plot_io(content)
    elif format == 'mat':
        from mlprodict.plotting.text_plot import onnx_text_plot
        code = onnx_text_plot(content)
    elif format == 'raw':
        code = str(content)
    elif format == 'tree':
        from mlprodict.plotting.plotting import onnx_text_plot_tree
        rows = []
        for node in content.graph.node:
            if node.op_type.startswith("TreeEnsemble"):
                rows.append(f'Node type={node.op_type!r} name={node.name!r}')
                rows.append(onnx_text_plot_tree(node))
        code = "\n".join(rows)
    else:
        raise ValueError(  # pragma: no cover
            f"Unknown format {format!r}.")

    if output not in ('', None):
        with open(output, "w", encoding="utf-8") as f:
            f.write(code)
    else:
        fLOG(code)  # pragma: no cover



"""
@file
@brief Computes shape inference for element wise operators with one input.
"""
import numpy
from .shape_excs import ShapeInferenceException
from .shape_result import OnnxKind


def _element_unary(known_shapes, node, dtype=None, one_input=True):
    """
    Infers shape for an element wise operator.
    The function returns but updates *known_shapes*.

    :param known_shapes: known shapes
    :param node: Onnx node
    :param dtype: None to keep the same type as input,
        not None to change it
    :param one_input: check there is only one input
    :return: updated or not
    """
    if one_input and len(node.input) != 1:
        raise ShapeInferenceException(  # pragma: no cover
            f"Node {node.name!r} must have one input not {len(node.input)}.")
    x = known_shapes[node.input[0]]
    if x.mtype != OnnxKind.Tensor:
        raise ShapeInferenceException(  # pragma: no cover
            f"Result {x!r} must be a tensor.")
    if dtype is None:
        return known_shapes.update(node.output[0], x.copy())
    cp = x.copy()
    cp.dtype = dtype
    return known_shapes.update(node.output[0], cp)


def shape_abs(known_shapes, node):
    "Infers shape for operator Abs."
    return _element_unary(known_shapes, node)


def shape_acos(known_shapes, node):
    "Infers shape for operator Acos."
    return _element_unary(known_shapes, node)


def shape_acosh(known_shapes, node):
    "Infers shape for operator Acosh."
    return _element_unary(known_shapes, node)


def shape_asin(known_shapes, node):
    "Infers shape for operator Asin."
    return _element_unary(known_shapes, node)


def shape_asinh(known_shapes, node):
    "Infers shape for operator Asinh."
    return _element_unary(known_shapes, node)


def shape_atan(known_shapes, node):
    "Infers shape for operator Atan."
    return _element_unary(known_shapes, node)


def shape_atanh(known_shapes, node):
    "Infers shape for operator Atanh."
    return _element_unary(known_shapes, node)


def shape_castlike(known_shapes, node):
    "Infers shape for operator CastLike."
    x = known_shapes[node.input[0]]
    if x.mtype != OnnxKind.Tensor:
        raise ShapeInferenceException(  # pragma: no cover
            f"Result {x!r} must be a tensor.")
    y = known_shapes[node.input[1]]
    if y.mtype != OnnxKind.Tensor:
        raise ShapeInferenceException(  # pragma: no cover
            f"Result {y!r} must be a tensor.")
    cp = x.copy()
    cp.dtype = y.dtype
    return known_shapes.update(node.output[0], cp)


def shape_ceil(known_shapes, node):
    "Infers shape for operator Ceil."
    return _element_unary(known_shapes, node)


def shape_celu(known_shapes, node):
    "Infers shape for operator Celu."
    return _element_unary(known_shapes, node)


def shape_clip(known_shapes, node):
    "Infers shape for operator Clip."
    return _element_unary(known_shapes, node, one_input=False)


def shape_cos(known_shapes, node):
    "Infers shape for operator Cos."
    return _element_unary(known_shapes, node)


def shape_cosh(known_shapes, node):
    "Infers shape for operator Cosh."
    return _element_unary(known_shapes, node)


def shape_elu(known_shapes, node):
    "Infers shape for operator Elu."
    return _element_unary(known_shapes, node)


def shape_erf(known_shapes, node):
    "Infers shape for operator Erf."
    return _element_unary(known_shapes, node)


def shape_exp(known_shapes, node):
    "Infers shape for operator Exp."
    return _element_unary(known_shapes, node)


def shape_floor(known_shapes, node):
    "Infers shape for operator Floor."
    return _element_unary(known_shapes, node)


def shape_hardmax(known_shapes, node):
    "Infers shape for operator Hardmax."
    return _element_unary(known_shapes, node)


def shape_hardsigmoid(known_shapes, node):
    "Infers shape for operator HardSigmoid."
    return _element_unary(known_shapes, node)


def shape_identity(known_shapes, node):
    "Infers shape for operator Identity."
    return _element_unary(known_shapes, node)


def shape_isnan(known_shapes, node):
    "Infers shape for operator IsNan."
    return _element_unary(known_shapes, node, numpy.bool_)


def shape_isinf(known_shapes, node):
    "Infers shape for operator IsInf."
    return _element_unary(known_shapes, node, numpy.bool_)


def shape_leakyrelu(known_shapes, node):
    "Infers shape for operator LeakyRelu."
    return _element_unary(known_shapes, node)


def shape_log(known_shapes, node):
    "Infers shape for operator Log."
    return _element_unary(known_shapes, node)


def shape_logsoftmax(known_shapes, node):
    "Infers shape for operator LogSoftmax."
    return shape_softmax(known_shapes, node)


def shape_neg(known_shapes, node):
    "Infers shape for operator Neg."
    return _element_unary(known_shapes, node)


def shape_not(known_shapes, node):
    "Infers shape for operator Not."
    x = known_shapes[node.input[0]]
    if x.dtype != numpy.bool_:
        raise ShapeInferenceException(
            f"Unexpected input type for operator Not {x.dtype!r} (must be bool).")
    return _element_unary(known_shapes, node)


def shape_reciprocal(known_shapes, node):
    "Infers shape for operator Reciprocal."
    return _element_unary(known_shapes, node)


def shape_relu(known_shapes, node):
    "Infers shape for operator Relu."
    return _element_unary(known_shapes, node)


def shape_round(known_shapes, node):
    "Infers shape for operator Round."
    return _element_unary(known_shapes, node)


def shape_selu(known_shapes, node):
    "Infers shape for operator Selu."
    return _element_unary(known_shapes, node)


def shape_shrink(known_shapes, node):
    "Infers shape for operator Shrink."
    return _element_unary(known_shapes, node)


def shape_sigmoid(known_shapes, node):
    "Infers shape for operator Sigmoid."
    return _element_unary(known_shapes, node)


def shape_sign(known_shapes, node):
    "Infers shape for operator Sigmoid."
    return _element_unary(known_shapes, node)


def shape_sin(known_shapes, node):
    "Infers shape for operator Sin."
    return _element_unary(known_shapes, node)


def shape_sinh(known_shapes, node):
    "Infers shape for operator Sinh."
    return _element_unary(known_shapes, node)


def shape_softmax(known_shapes, node):
    "Infers shape for operator Softmax."
    return _element_unary(known_shapes, node)


def shape_softplus(known_shapes, node):
    "Infers shape for operator Softplus."
    return _element_unary(known_shapes, node)


def shape_softsign(known_shapes, node):
    "Infers shape for operator Softsign."
    return _element_unary(known_shapes, node)


def shape_sqrt(known_shapes, node):
    "Infers shape for operator Sqrt."
    return _element_unary(known_shapes, node)


def shape_tan(known_shapes, node):
    "Infers shape for operator Tan."
    return _element_unary(known_shapes, node)


def shape_tanh(known_shapes, node):
    "Infers shape for operator Tanh."
    return _element_unary(known_shapes, node)


def shape_thresholdedrelu(known_shapes, node):
    "Infers shape for operator ThresholdedRelu."
    return _element_unary(known_shapes, node)


def shape_trilu(known_shapes, node):
    "Infers shape for operator Trilu."
    return _element_unary(known_shapes, node)



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from ._op import OpRun


class CategoryMapper(OpRun):

    atts = {'cats_int64s': numpy.empty(0, dtype=numpy.int64),
            'cats_strings': numpy.empty(0, dtype=numpy.str_),
            'default_int64': -1,
            'default_string': b'',
            }

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       expected_attributes=CategoryMapper.atts,
                       **options)
        if len(self.cats_int64s) != len(self.cats_strings):
            raise RuntimeError(  # pragma: no cover
                "Lengths mismatch between cats_int64s (%d) and "
                "cats_strings (%d)." % (
                    len(self.cats_int64s), len(self.cats_strings)))
        self.int2str_ = {}
        self.str2int_ = {}
        for a, b in zip(self.cats_int64s, self.cats_strings):
            be = b.decode('utf-8')
            self.int2str_[a] = be
            self.str2int_[be] = a

    def _run(self, x, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        if x.dtype == numpy.int64:
            xf = x.ravel()
            res = [self.int2str_.get(xf[i], self.default_string)
                   for i in range(0, xf.shape[0])]
            return (numpy.array(res).reshape(x.shape), )

        xf = x.ravel()
        res = numpy.empty((xf.shape[0], ), dtype=numpy.int64)
        for i in range(0, res.shape[0]):
            res[i] = self.str2int_.get(xf[i], self.default_int64)
        return (res.reshape(x.shape), )



"""
@file
@brief Design to implement graph as parameter.

.. versionadded:: 0.8
"""
import logging
import numpy
from .onnx_variable import OnnxVar
from .xop import loadop


logger = logging.getLogger('xop')


class AttributeGraph:
    """
    Class wrapping a function to make it simple as
    a parameter.

    :param fct: function taking the list of inputs defined
        as @see cl OnnxVar, the function returns an @see cl OnnxVar
    :param inputs: list of input as @see cl OnnxVar

    .. versionadded:: 0.8
    """

    def __init__(self, fct, *inputs):
        logger.debug('AttributeGraph(%r, %d in)', type(fct), len(inputs))
        if isinstance(fct, numpy.ndarray) and len(inputs) == 0:
            self.cst = fct
            fct = None
        else:
            self.cst = None
        self.fct = fct
        self.inputs = inputs
        self.alg_ = None

    def __repr__(self):
        "usual"
        return f"{self.__class__.__name__}(...)"

    def _graph_guess_dtype(self, i, var):
        """
        Guesses the graph inputs.

        :param i: attribute index (integer)
        :param var: the input (@see cl OnnxVar)
        :return: input type
        """
        dtype = var._guess_dtype(None)
        if dtype is None:
            dtype = numpy.float32

        input_name = 'graph_%d_%d' % (id(self), i)
        return OnnxVar(input_name, dtype=dtype)

    def to_algebra(self, op_version=None):
        """
        Converts the variable into an operator.
        """
        if self.alg_ is not None:
            return self.alg_

        logger.debug('AttributeGraph.to_algebra(op_version=%r)',
                     op_version)
        if self.cst is not None:
            OnnxIdentity = loadop('Identity')
            self.alg_ = OnnxIdentity(self.cst, op_version=op_version)
            self.alg_inputs_ = None
            logger.debug('AttributeGraph.to_algebra:end:1:%r', type(self.alg_))
            return self.alg_

        new_inputs = [self._graph_guess_dtype(i, inp)
                      for i, inp in enumerate(self.inputs)]
        self.alg_inputs_ = new_inputs
        vars = [v[1] for v in new_inputs]
        var = self.fct(*vars)
        if not isinstance(var, OnnxVar):
            raise RuntimeError(  # pragma: no cover
                f"var is not from type OnnxVar but {type(var)!r}.")

        self.alg_ = var.to_algebra(op_version=op_version)
        logger.debug('AttributeGraph.to_algebra:end:2:%r', type(self.alg_))
        return self.alg_


class OnnxVarGraph(OnnxVar):
    """
    Overloads @see cl OnnxVar to handle graph attribute.

    :param inputs: variable name or object
    :param op: :epkg:`ONNX` operator
    :param select_output: if multiple output are returned by
        ONNX operator *op*, it takes only one specifed by this
        argument
    :param dtype: specifies the type of the variable
        held by this class (*op* is None) in that case
    :param fields: list of attributes with the graph type
    :param kwargs: addition argument to give operator *op*

    .. versionadded:: 0.8
    """

    def __init__(self, *inputs, op=None, select_output=None,
                 dtype=None, **kwargs):
        OnnxVar.__init__(
            self, *inputs, op=op, select_output=select_output,
            dtype=dtype, **kwargs)

    def to_algebra(self, op_version=None):
        """
        Converts the variable into an operator.
        """
        if self.alg_ is not None:
            return self.alg_

        logger.debug('OnnxVarGraph.to_algebra(op_version=%r)',
                     op_version)
        # Conversion of graph attributes from InputGraph
        # ONNX graph.
        updates = dict()
        self.alg_hidden_var_ = {}
        self.alg_hidden_var_inputs = {}
        for att, var in self.onnx_op_kwargs.items():
            if not isinstance(var, AttributeGraph):
                continue
            alg = var.to_algebra(op_version=op_version)
            if var.alg_inputs_ is None:
                onnx_inputs = []
            else:
                onnx_inputs = [i[0] for i in var.alg_inputs_]
            onx = alg.to_onnx(onnx_inputs, target_opset=op_version)
            updates[att] = onx.graph
            self.alg_hidden_var_[id(var)] = var
            self.alg_hidden_var_inputs[id(var)] = onnx_inputs
        self.onnx_op_kwargs_before = {
            k: self.onnx_op_kwargs[k] for k in updates}
        self.onnx_op_kwargs.update(updates)
        self.alg_ = OnnxVar.to_algebra(self, op_version=op_version)
        logger.debug('OnnxVarGraph.to_algebra:end:%r', type(self.alg_))
        return self.alg_


class if_then_else(AttributeGraph):
    """
    Overloads class @see cl OnnxVarGraph.
    """

    def __init__(self, fct, *inputs):
        AttributeGraph.__init__(self, fct, *inputs)



"""
@file
@brief Automates the generation of operators for the
documentation for the Xop API.

.. versionadded:: 0.9
"""
import os
import textwrap
import importlib
import inspect
import onnx
import onnx.defs
from onnx.backend.test.case.base import _Exporter
from onnx.onnx_cpp2py_export.defs import SchemaError  # pylint: disable=E1101,E0611,E0401
from onnx.defs import OpSchema


def _get_doc_template():
    try:
        from jinja2 import Template
    except ImportError:  # pragma no cover
        class Template:
            "Docstring template"

            def __init__(self, *args):
                pass

            def render(self, **context):
                "render"
                schemas = context['schemas']
                rows = []
                for sch in schemas:
                    doc = sch.doc or ''
                    name = sch.name
                    if name is None:
                        raise RuntimeError("An operator must have a name.")
                    rows.extend([name, "=" * len(name),
                                 "", doc, ""])
                return "\n".join(rows)

    return Template(textwrap.dedent("""
        {% for sch in schemas %}

        .. tag-diff-insert.

        .. _l-onnx-op{{sch.domain.lower().replace(".", "-")}}-{{sch.name.lower()}}-{{str(sch.since_version)}}:

        {{format_name_with_domain(sch)}}
        {{'=' * len(format_name_with_domain(sch))}}

        **Version**

        * **name**: `{{sch.name}} (GitHub) <{{build_doc_url(sch)}}{{sch.name}}>`_
        * **domain**: **{% if sch.domain == '' %}main{% else %}{{sch.domain}}{% endif %}**
        * **since_version**: **{{sch.since_version}}**
        * **function**: {{sch.has_function}}
        * **support_level**: {{sch.support_level}}
        * **shape inference**: {{sch.has_type_and_shape_inference_function}}

        {% if sch.support_level == OpSchema.SupportType.EXPERIMENTAL %}
        No versioning maintained for experimental ops.
        {% else %}
        This version of the operator has been {% if
        sch.deprecated %}deprecated{% else %}available{% endif %}
        **since version {{sch.since_version}}{% if
        sch.domain %} of domain {{sch.domain}}{% endif %}**.
        {% if len(sch.versions) > 1 %}
        Other versions of this operator:
        {% for v in sch.version[:-1] %} {{v}} {% endfor %}
        {% endif %}
        {% endif %}

        **Summary**

        {{process_documentation(sch.doc)}}

        {% if sch.attributes %}
        **Attributes**

        {% for _, attr in sorted(sch.attributes.items()) %}* **{{attr.name}}**{%
          if attr.required %} (required){% endif %}:
        {{text_wrap(attr.description, 2)}} {%
          if attr.default_value %}{{clean_default_value(attr.default_value)}}{%
          endif %}
        {% endfor %}
        {% endif %}

        {% if sch.inputs %}
        **Inputs**

        {% if sch.min_input != sch.max_input %}Between {{sch.min_input
        }} and {{sch.max_input}} inputs.
        {% endif %}
        {% for ii, inp in enumerate(sch.inputs) %}
        * **{{getname(inp, ii)}}**{{format_option(inp)}} - **{{inp.typeStr}}**:
        {{text_wrap(inp.description, 2)}}{% endfor %}
        {% endif %}

        {% if sch.outputs %}
        **Outputs**

        {% if sch.min_output != sch.max_output %}Between {{sch.min_output
        }} and {{sch.max_output}} outputs.
        {% endif %}
        {% for ii, out in enumerate(sch.outputs) %}
        * **{{getname(out, ii)}}**{{format_option(out)}} - **{{out.typeStr}}**:
        {{text_wrap(out.description, 2)}}{% endfor %}
        {% endif %}

        {% if sch.type_constraints %}
        **Type Constraints**

        {% for ii, type_constraint in enumerate(sch.type_constraints)
        %}* {{get_constraint(type_constraint, ii)}}:
        {{text_wrap(type_constraint.description, 2)}}
        {% endfor %}
        {% endif %}

        {% if get_onnx_example and is_last_schema(sch): %}
        **Examples**

        {% for example, code in get_onnx_example(sch.name).items(): %}
        **{{ example }}**

        ::

        {{ format_example(code) }}

        {% endfor %}
        {% endif %}

        {% endfor %}
    """))


_template_operator = _get_doc_template()
__get_all_schemas_with_history = None


def _populate__get_all_schemas_with_history():
    res = {}
    for schema in onnx.defs.get_all_schemas_with_history():
        domain = schema.domain
        version = schema.since_version
        name = schema.name
        if domain not in res:
            res[domain] = {}
        if name not in res[domain]:
            res[domain][name] = {}
        res[domain][name][version] = schema

    try:
        import onnxruntime.capi.onnxruntime_pybind11_state as rtpy
    except ImportError:  # pragma: no cover
        rtpy = None

    if rtpy is not None:
        # If onnxruntime is available, it is being populated with these operators as well.
        from .xop import _CustomSchema
        try:
            get_schemas = rtpy.get_all_operator_schema
        except AttributeError:
            # onnxruntime must be compiled with flag --gen_doc.
            # a local copy is retrieved.
            from .xop import _get_all_operator_schema
            get_schemas = _get_all_operator_schema
        for op in get_schemas():
            sch = _CustomSchema(op)
            domain, name = sch.domain, sch.name
            if domain in res and name in res[domain]:
                # already handled
                continue
            version = sch.since_version
            if domain not in res:
                res[domain] = {}
            if name not in res[domain]:
                res[domain][name] = {}
            res[domain][name][version] = sch

    return res


def _get_all_schemas_with_history():
    global __get_all_schemas_with_history  # pylint: disable=W0603
    if __get_all_schemas_with_history is None:
        __get_all_schemas_with_history = _populate__get_all_schemas_with_history()
    return __get_all_schemas_with_history


def get_domain_list():
    """
    Returns the list of available domains.
    """
    return list(sorted(set(map(lambda s: s.domain,
                               onnx.defs.get_all_schemas_with_history()))))


def get_operator_schemas(op_name, version=None, domain=None):
    """
    Returns all schemas mapped to an operator name.

    :param op_name: name of the operator
    :param version: version
    :param domain: domain
    :return: list of schemas
    """
    if version == 'last' and op_name is not None:
        if domain is not None:
            return [onnx.defs.get_schema(op_name, domain=domain)]
    all_schemas = _get_all_schemas_with_history()
    if domain is None:
        domains = []
        for dom, ops in all_schemas.items():
            if op_name is None or op_name in ops:
                domains.append(dom)
    else:
        domains = [domain]

    # schemas
    sch = []
    for dom in domains:
        ops = all_schemas[dom]
        if op_name is None:
            for op, v in ops.items():
                if version is None:
                    sch.extend(v.values())
                elif version == 'last' and (dom == '' or 'onnx' in dom):
                    try:
                        sch.append(onnx.defs.get_schema(op, domain=dom))
                    except SchemaError:  # pragma: no cover
                        sch.append(v[max(v)])
                elif version == 'last':
                    sch.append(v[max(v)])
                else:
                    sch.append(v[version])
        elif op_name in ops:
            if version is None:
                sch.extend(ops[op_name].values())
            elif version in ops[op_name]:
                sch.append(ops[op_name][version])

    # sort
    vals = [(s.domain, s.name, -s.since_version, s) for s in sch]
    vals.sort()
    return [v[-1] for v in vals]


def get_rst_doc(op_name=None, domain=None, version='last', clean=True,
                diff=False, example=False):
    """
    Returns a documentation in RST format
    for all :class:`OnnxOperator`.

    :param op_name: operator name of None for all
    :param domain: domain
    :param version: version, None for all, `'last'` for the most recent one
    :param clean: clean empty lines
    :param diff: highlights differences between two versions
    :param example: add example to the documentation
    :return: string

    The function relies on module :epkg:`jinja2` or replaces it
    with a simple rendering if not present.
    """
    from ..onnx_tools.onnx2py_helper import _var_as_dict
    schemas = get_operator_schemas(op_name, domain=domain, version=version)

    # from onnx.backend.sample.ops import collect_sample_implementations
    # from onnx.backend.test.case import collect_snippets
    # SNIPPETS = collect_snippets()
    # SAMPLE_IMPLEMENTATIONS = collect_sample_implementations()
    def format_name_with_domain(sch):
        if version == 'last':
            if sch.domain:
                return f'{sch.name} ({sch.domain})'
            return sch.name
        if sch.domain:
            return f'{sch.name} - {sch.since_version} ({sch.domain})'
        return '%s - %d' % (sch.name, sch.since_version)

    def format_option(obj):
        opts = []
        if OpSchema.FormalParameterOption.Optional == obj.option:
            opts.append('optional')
        elif OpSchema.FormalParameterOption.Variadic == obj.option:
            opts.append('variadic')
        if getattr(obj, 'isHomogeneous', False):
            opts.append('heterogeneous')
        if opts:
            return f" ({', '.join(opts)})"
        return ""

    def format_example(code):
        code = textwrap.indent(code, '    ')
        return code

    def get_constraint(const, ii):
        if const.type_param_str:
            name = const.type_param_str
        else:
            name = str(ii)
        name = f"**{name}** in ("
        if const.allowed_type_strs:
            text = ",\n  ".join(sorted(const.allowed_type_strs))
            name += "\n  " + text + "\n  )"
        return name

    def getname(obj, i):
        name = obj.name
        if len(name) == 0:
            return str(i)
        return name

    def process_documentation(doc):
        if doc is None:
            doc = ''
        if not isinstance(doc, str):
            raise TypeError(  # pragma: no cover
                f"doc must be a string not {type(doc)!r} - {doc + 42!r}.")
        doc = textwrap.dedent(doc)
        main_docs_url = "https://github.com/onnx/onnx/blob/master/"
        rep = {
            '[the doc](IR.md)': '`ONNX <{0}docs/IR.md>`_',
            '[the doc](Broadcasting.md)':
                '`Broadcasting in ONNX <{0}docs/Broadcasting.md>`_',
            '<dl>': '',
            '</dl>': '',
            '<dt>': '* ',
            '<dd>': '  ',
            '</dt>': '',
            '</dd>': '',
            '<tt>': '``',
            '</tt>': '``',
            '<br>': '\n',
        }
        for k, v in rep.items():
            doc = doc.replace(k, v.format(main_docs_url))
        move = 0
        lines = []
        for line in doc.split('\n'):
            if line.startswith("```"):
                if move > 0:
                    move -= 4
                    lines.append("\n")
                else:
                    lines.append("::\n")
                    move += 4
            elif move > 0:
                lines.append(" " * move + line)
            else:
                lines.append(line)
        return "\n".join(lines)

    def build_doc_url(sch):
        doc_url = "https://github.com/onnx/onnx/blob/main/docs/Operators"
        if "ml" in sch.domain:
            doc_url += "-ml"
        doc_url += ".md"
        doc_url += "#"
        if sch.domain not in (None, '', 'ai.onnx'):
            doc_url += sch.domain + "."
        return doc_url

    def clean_default_value(value):
        dvar = _var_as_dict(value)
        if 'value' in dvar:
            v = dvar['value']
            if isinstance(v, bytes):
                return f"Default value is ``'{v.decode('ascii')}'``."
            return f"Default value is ``{v}``."
        else:
            res = str(value).replace('\n', ' ').strip()
            if len(res) > 0:
                return f"Default value is ``{res}``."
        return ""

    def text_wrap(text, indent):
        s = ' ' * indent
        lines = textwrap.wrap(text, initial_indent=s, subsequent_indent=s)
        return '\n'.join(lines)

    fnwd = format_name_with_domain
    tmpl = _template_operator
    docs = tmpl.render(schemas=schemas, OpSchema=OpSchema,
                       len=len, getattr=getattr, sorted=sorted,
                       format_option=format_option,
                       get_constraint=get_constraint,
                       getname=getname, enumerate=enumerate,
                       format_name_with_domain=fnwd,
                       process_documentation=process_documentation,
                       build_doc_url=build_doc_url, text_wrap=text_wrap,
                       str=str, clean_default_value=clean_default_value,
                       get_onnx_example=get_onnx_example if example else None,
                       format_example=format_example,
                       is_last_schema=is_last_schema)
    if diff:
        lines = docs.split('\n')
        new_lines = ['']
        for line in lines:
            line = line.rstrip('\r\t ')
            if len(line) == 0 and len(new_lines[-1]) == 0:
                continue
            new_lines.append(line)
        docs = '\n'.join(new_lines)
        docs = _insert_diff(docs, '.. tag-diff-insert.')

    if clean:
        lines = docs.split('\n')
        new_lines = ['']
        for line in lines:
            line = line.rstrip('\r\t ')
            if len(line) == 0 and len(new_lines[-1]) == 0:
                continue
            new_lines.append(line)
        docs = '\n'.join(new_lines)

    return docs


def _insert_diff(docs, split='.. tag-diff-insert.'):
    """
    Splits a using `split`, insert HTML differences between pieces.
    The function relies on package :epkg:`pyquickhelper`.
    """
    spl = docs.split(split)
    if len(spl) <= 1:
        return docs

    from pyquickhelper.texthelper.edit_text_diff import (
        edit_distance_text, diff2html)

    pieces = [spl[0]]
    for i in range(1, len(spl)):
        spl1 = spl[i - 1].strip('\n ')
        spl2 = spl[i].strip('\n ')
        spl1 = spl1.split('**Examples**')[0].replace('`', '')
        spl2 = spl2.split('**Examples**')[0].replace('`', '')
        spl1 = spl1.split('**Summary**')[-1].strip('\n ')
        spl2 = spl2.split('**Summary**')[-1].strip('\n ')
        if len(spl1) < 5 or len(spl2) < 5:
            pieces.append(spl[i])
            continue

        _, aligned, final = edit_distance_text(  # pylint: disable=W0632
            spl2, spl1, threshold=0.5)
        ht = diff2html(spl2, spl1, aligned, final, two_columns=True)
        ht = ht.replace(">``<", "><")
        ht = '    ' + '\n    '.join(ht.split('\n'))
        pieces.extend(['', '**Differences**', '', '.. raw:: html',
                       '', ht, '', spl[i]])

    return '\n'.join(pieces)


def get_onnx_example(op_name):
    """
    Retrieves examples associated to one operator
    stored in onnx packages.

    :param op_name: operator name
    :param fmt: rendering format
    :return: dictionary
    """
    module = f'onnx.backend.test.case.node.{op_name.lower()}'
    try:
        mod = importlib.import_module(module)
    except ImportError:
        return {}
    results = {}
    for v in mod.__dict__.values():
        if not isinstance(v, _Exporter):
            continue
        code_cls = inspect.getsource(v)
        codes = code_cls.split('@staticmethod')
        for me in v.__dict__:
            if not me.startswith('export_'):
                continue
            sub = f' {me}()'
            found = None
            for code in codes:
                if sub in code:
                    found = code
            if found is None:
                raise RuntimeError(  # pragma: no cover
                    f"Unable to find {sub!r} in\n{code_cls}")
            found = textwrap.dedent(found)
            lines = found.split('\n')
            first = 0
            for i in range(len(lines)):  # pylint: disable=C0200
                if lines[i].startswith('def '):
                    first = i + 1
            found = textwrap.dedent('\n'.join(lines[first:]))
            results[me[len('export_'):]] = found
    return results


def is_last_schema(sch):
    """
    Tells if this is the most recent schema for this operator.

    :param sch: schema
    :return: True
    """
    try:
        last = onnx.defs.get_schema(sch.name, domain=sch.domain)
    except SchemaError:  # pragma: no cover
        # raise RuntimeError(
        #     "Unable to find schema for operator %r and domain %r."
        #     "" % (sch.name, sch.domain))
        return True
    return last.since_version == sch.since_version


def onnx_documentation_folder(folder, ops=None, title='ONNX operators',
                              fLOG=None):
    """
    Creates documentation in a folder for all known
    ONNX operators or a subset.

    :param folder: folder where to write the documentation
    :param ops: None for all operators or a subset of them
    :param title: index title
    :param fLOG: logging function
    :return: list of creates files
    """
    all_schemas = _get_all_schemas_with_history()
    if not os.path.exists(folder):
        os.makedirs(folder)
    index = ['', title, '=' * len(title), '', '.. contents::',
             '    :local:', '']
    pages = []
    tables_domain_pages = []

    if ops is not None:
        ops = set(ops)
    for dom in sorted(all_schemas):
        sdom = 'main' if dom == '' else dom

        index_dom = [sdom, '+' * len(sdom), '', '.. toctree::',
                     '    :maxdepth: 1', '']

        table_dom = ["", f".. _l-table-operator-{sdom.replace('.', '-')}:", "",
                     f"operator table for domain {sdom}"]
        table_dom.extend(["=" * len(table_dom[-1]), ""])
        table_dom.extend([f".. list-table:: operators for domain {sdom}",
                          "    :widths: 10 10",
                          "    :header-rows: 1",
                          "",
                          "    * - operator",
                          "      - versions"])

        sub = all_schemas[dom]
        do = []
        if ops is None:
            do.extend(sub)
        else:
            inter = set(sub).intersection(ops)
            if len(inter) == 0:
                continue
            do.extend(sorted(inter))
        if len(do) == 0:
            continue

        for op in sorted(do):
            if fLOG is not None:
                fLOG(  # pragma: no cover
                    f'generate page for onnx {dom!r} - {op!r}')
            page_name = f"onnx_{dom.replace('.', '')}_{op}"
            index_dom.append(f'    {page_name}')
            doc = get_rst_doc(op, domain=dom, version=None, example=True,
                              diff=True)
            if dom == '':
                main = op
            else:
                main = f'{dom} - {op}'
            rows = ['', f'.. _l-onnx-doc{dom}-{op}:', '',
                    '=' * len(main), main, '=' * len(main), '',
                    '.. contents::', '    :local:', '', doc]

            full = os.path.join(folder, page_name + '.rst')
            with open(full, 'w', encoding='utf-8') as f:
                f.write("\n".join(rows))
            pages.append(full)

            # table
            schemas = get_operator_schemas(op, domain=dom, version=None)
            links = []
            for sch in schemas:
                link = (
                    ':ref:`{sver} <l-onnx-op{lname_}-{lname}-{sver}>`').format(
                        sver=str(sch.since_version), lname=sch.name.lower(),
                        lname_=sch.domain.lower().replace(".", "-"))
                links.append(link)
            table_dom.extend([f"    * - {op}",
                              f"      - {', '.join(links)}"])

        sdom_clean = sdom.replace('.', '_')
        page_name = os.path.join(folder, f'table_{sdom_clean}.rst')
        tables_domain_pages.append(f'table_{sdom_clean}')
        pages.append(page_name)
        with open(page_name, "w", encoding="utf-8") as f:
            f.write("\n".join(table_dom))

        index.extend(index_dom)
        index.append('')

    # adding pages
    index.extend(["", "Tables", "++++++", "",
                 ".. toctree::", "    :maxdepth: 1", ""])
    for page in tables_domain_pages:
        index.append(f"    {page}")
    index.append('')

    # creating a big index
    page_name = os.path.join(folder, 'index.rst')
    with open(page_name, 'w', encoding='utf-8') as f:
        f.write('\n'.join(index))
    pages.append(page_name)
    return pages



"""
@file
@brief Easier API to build onnx graphs. Inspired from :epkg:`skl2onnx`.

.. versionadded:: 0.9
"""
import logging
import numpy
from .xop import OnnxOperator, OnnxOperatorFunction
from .xop_variable import NodeResultName, Variable


logger = logging.getLogger('xop')


class OnnxSubOnnx(OnnxOperator):
    """
    This operator is used to insert existing ONNX into
    the ONNX graph being built.
    """

    domain = 'mlprodict'
    since_version = 1
    expected_inputs = None
    expected_outputs = None
    input_range = [1, 1e9]
    output_range = [1, 1e9]
    op_type = 'SubOnnx'
    domain = 'mlprodict.xop'

    def __init__(self, model, *inputs, output_names=None):
        logger.debug("SubOnnx(ONNX, %d in, output_names=%r)",
                     len(inputs), output_names)
        if model is None:
            raise ValueError("Model cannot be None.")  # pragma: no cover
        if len(inputs) > len(model.graph.input):
            raise RuntimeError(  # pragma: no cover
                "Unexpected number of inputs %r > expected %r." % (
                    len(inputs), len(model.graph.input)))
        if (output_names is not None and
                len(output_names) != len(model.graph.output)):
            raise RuntimeError(  # pragma: no cover
                "Unexpected number of outputs %r != expected %r." % (
                    len(output_names), len(model.graph.output)))
        if len(inputs) == 0:
            if hasattr(model, 'graph'):
                inputs = [Variable(i.name, i.type.tensor_type)
                          for i in model.graph.input]
            else:
                inputs = [Variable(n) for n in model.input]
        OnnxOperator.__init__(self, *inputs, output_names=output_names)
        if self.output_names is None and self.expected_outputs is None:
            if hasattr(model, 'graph'):
                self.expected_outputs = [
                    (i.name, i.type.tensor_type)
                    for i in model.graph.output]
            else:
                self.expected_outputs = [(n, None) for n in model.output]
        self.model = model

    @property
    def input_names(self):
        "Returns the input names."
        return ([i.name for i in self.model.graph.input]
                if hasattr(self.model, 'graph') else list(self.model.input))

    def __repr__(self):
        "usual"
        atts = {}
        for att in ['output_names']:
            value = getattr(self, att, None)
            if value is not None:
                atts[att] = value
        atts.update(self.kwargs)
        msg = ", ".join(f"{k}={v!r}" for k, v in atts.items())
        if len(atts) > 0:
            msg = ", " + msg
        return f"{self.__class__.__name__}(...{msg})"

    def add_to(self, builder):
        """
        Adds to graph builder.

        :param builder: instance of @see cl _GraphBuilder,
            it must have a method `add_node`
        """
        logger.debug("SubOnnx.add_to(builder)")
        inputs = builder.get_input_names(self, self.inputs)
        n_outputs = len(self.model.graph.output)
        outputs = [builder.get_unique_output_name(NodeResultName(self, i))
                   for i in range(n_outputs)]

        mapped_names = {}

        # adding initializers
        for init in self.model.graph.initializer:
            new_name = builder.get_unique_name(init.name, reserved=False)
            mapped_names[init.name] = new_name
            builder.add_initializer(new_name, init)

        # linking inputs
        for inp, name in zip(self.model.graph.input, inputs):
            new_name = builder.get_unique_name(inp.name, reserved=False)
            mapped_names[inp.name] = new_name
            builder.add_node(
                'Identity', builder.get_unique_name(
                    '_sub_' + name, reserved=False),
                [name], [new_name])

        # adding nodes
        for node in self.model.graph.node:
            new_inputs = []
            for i in node.input:
                if i not in mapped_names:
                    raise RuntimeError(  # pragma: no cover
                        f"Unable to find input {i!r} in {mapped_names!r}.")
                new_inputs.append(mapped_names[i])
            new_outputs = []
            for o in node.output:
                new_name = builder.get_unique_name(o, reserved=False)
                mapped_names[o] = new_name
                new_outputs.append(new_name)

            atts = {}
            for att in node.attribute:
                atts[att.name] = OnnxOperatorFunction.attribute_to_value(att)

            builder.add_node(
                node.op_type,
                builder.get_unique_name('_sub_' + node.name, reserved=False),
                new_inputs, new_outputs, domain=node.domain, **atts)

        # linking outputs
        for out, name in zip(self.model.graph.output, outputs):
            builder.add_node(
                'Identity', builder.get_unique_name(
                    '_sub_' + out.name, reserved=False),
                [mapped_names[out.name]], [name])

    def to_onnx_this(self, evaluated_inputs):
        """
        Returns the ONNX graph.

        :param evaluated_inputs: unused
        :return: ONNX graph
        """
        return self.model


class OnnxSubEstimator(OnnxSubOnnx):
    """
    This operator is used to call the converter of a model
    to insert the node coming from the conversion into a
    bigger ONNX graph. It supports model from :epkg:`scikit-learn`
    using :epkg:`sklearn-onnx`.

    :param model: model to convert
    :param inputs: inputs
    :param op_version: targetted opset
    :param options: to rewrite the options used to convert the model
    :param initial_types: the implementation may be wrong in guessing
        the input types of the model, this parameter can be used
        to overwrite them, usually a dictionary
        `{ input_name: numpy array as an example }`
    :param kwargs: any other parameters such as black listed or
        white listed operators
    """

    since_version = 1
    expected_inputs = None
    expected_outputs = None
    input_range = [1, 1e9]
    output_range = [1, 1e9]
    op_type = "SubEstimator"
    domain = 'mlprodict.xop'

    def __init__(self, model, *inputs, op_version=None,
                 output_names=None, options=None,
                 initial_types=None, **kwargs):
        logger.debug("OnnxSubEstimator(%r, %r, op_version=%r, "
                     "output_names=%r, initial_types=%r, options=%r, "
                     "kwargs=%r)", type(model), inputs, op_version,
                     output_names, initial_types, options, kwargs)
        if model is None:
            raise ValueError("Model cannot be None.")  # pragma: no cover
        onx = OnnxSubEstimator._to_onnx(
            model, inputs, op_version=op_version, options=options,
            initial_types=initial_types, **kwargs)
        OnnxSubOnnx.__init__(
            self, onx, *inputs, output_names=output_names)
        self.ml_model = model
        self.options = options
        self.initial_types = initial_types
        self.op_version = op_version

    def __repr__(self):
        "usual"
        atts = {}
        for att in ['op_version', 'output_names', 'options',
                    'initial_types']:
            value = getattr(self, att, None)
            if value is not None:
                atts[att] = value
        atts.update(self.kwargs)
        msg = ", ".join(f"{k}={v!r}" for k, v in atts.items())
        if len(atts) > 0:
            msg = ", " + msg
        return f"{self.__class__.__name__}({self.ml_model!r}{msg})"

    @staticmethod
    def _to_onnx(model, inputs, op_version=None, options=None,
                 initial_types=None, **kwargs):
        """
        Converts a model into ONNX and inserts it into an ONNX graph.

        :param model: a trained machine learned model
        :param inputs: inputs
        :param op_version: opset versions or None to use the latest one
        :param options: options to change the behaviour of the converter
        :param kwargs: additional parameters such as black listed or while listed
            operators
        :return: ONNX model

        The method currently supports models trained with
        :epkg:`scikit-learn`, :epkg:`xgboost`, :epkg`:lightgbm`.
        """
        from sklearn.base import BaseEstimator

        if isinstance(model, BaseEstimator):
            logger.debug("OnnxSubEstimator._to_onnx(%r, %r, op_version=%r "
                         "options=%r, initial_types=%r, kwargs=%r)",
                         type(model), inputs, op_version, options,
                         initial_types, kwargs)
            return OnnxSubEstimator._to_onnx_sklearn(
                model, inputs, op_version=op_version, options=options,
                initial_types=initial_types, **kwargs)
        raise RuntimeError(  # pragma: no cover
            f"Unable to convert into ONNX model type {type(model)!r}.")

    @staticmethod
    def _to_onnx_sklearn(model, inputs, op_version=None, options=None,
                         initial_types=None, **kwargs):
        """
        Converts a :epkg:`scikit-learn` model into ONNX
        and inserts it into an ONNX graph. The library relies on
        function @see fn to_onnx and library :epkg:`skearn-onnx`.

        :param model: a trained machine learned model
        :param inputs: inputs
        :param op_version: opset versions or None to use the latest one
        :param initial_types: if None, the input types are guessed from the
            inputs. The function converts into ONNX the previous
            node of the graph and tries to infer the initial_types
            with the little informations it has. It may not work.
            It is recommended to specify this parameter.
        :param options: options to change the behaviour of the converter
        :param kwargs: additional parameters such as black listed or while listed
            operators
        :return: ONNX model

        Default options is `{'zipmap': False}` for a classifier.
        """
        from ..onnx_conv.convert import to_onnx
        if options is None:
            from sklearn.base import ClassifierMixin
            if isinstance(model, ClassifierMixin):
                options = {'zipmap': False}
        if initial_types is None:
            # adding more information
            from skl2onnx.common.data_types import _guess_numpy_type  # delayed
            for i, n in enumerate(inputs):
                if not isinstance(n, Variable):
                    raise NotImplementedError(
                        "Inpput %d is not a variable but %r." % (i, type(n)))
            initial_types = [(n.name, _guess_numpy_type(n.dtype, n.shape))
                             for n in inputs]

        logger.debug("OnnxSubEstimator._to_onnx_sklearn(%r, %r, "
                     "op_version=%r, options=%r, initial_types=%r, "
                     "kwargs=%r)",
                     type(model), inputs, op_version, options,
                     initial_types, kwargs)

        if isinstance(initial_types, numpy.ndarray):
            if len(inputs) != 1:
                raise RuntimeError(  # pragma: no cover
                    "The model has %s inputs but only %d input are "
                    "described in 'initial_types'." % (
                        len(inputs), 1))
            X = initial_types
            initial_types = None
        elif len(inputs) != len(initial_types):
            raise RuntimeError(  # pragma: no cover
                "The model has %s inputs but only %d input are "
                "described in 'initial_types'." % (
                    len(inputs), len(initial_types)))
        else:
            X = None

        onx = to_onnx(model, X, initial_types=initial_types, options=options,
                      rewrite_ops=True, target_opset=op_version, **kwargs)
        return onx



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from ._op import OpRun


class DequantizeLinear(OpRun):

    atts = {'axis': 1}
    python_inputs = ['*inputs']

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       expected_attributes=DequantizeLinear.atts,
                       **options)

    def _run(self, *args, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        if len(args[1].shape) > 1:
            raise RuntimeError(  # pragma: no cover
                "Input 2 must be a vector or a number.")

        x_scale = args[2]
        if len(x_scale.shape) > 0 and x_scale.size == 1:
            x_scale = x_scale[0]
        if len(args) > 2:
            if x_scale.dtype != args[0].dtype:
                raise RuntimeError(  # pragma no cover
                    "Type mismatch {} != {} in DequantizeLinear.".format(
                        args[0].dtype, x_scale.dtype))

            if len(x_scale.shape) > 0:
                new_shape = [1 for s in args[0].shape]
                new_shape[self.axis] = len(x_scale)
                x = args[0].astype(numpy.float32) - x_scale.reshape(new_shape)
                y = x * args[1].reshape(new_shape)
            else:
                x = args[0].astype(numpy.float32) - x_scale
                y = x * args[1]
        elif len(args[1].shape) > 0:
            new_shape = [1 for s in args[0].shape]
            new_shape[self.axis] = len(x_scale)
            y = args[0].astype(numpy.float32) * x_scale.reshape(new_shape)
        else:
            y = args[0].astype(numpy.float32) * x_scale
        return (y.astype(numpy.float32), )



"""
@file
@brief Implements :epkg:`numpy` functions with onnx and a runtime.

.. versionadded:: 0.6
"""
import inspect
import logging
from typing import Any
import numpy
from ..onnx_tools.optim._main_onnx_optim import onnx_optimisations
from .onnx_version import FctVersion
from .onnx_numpy_annotation import get_args_kwargs
from .xop_variable import Variable
from .xop import OnnxOperator, OnnxOperatorTuple


logger = logging.getLogger('xop')


class OnnxNumpyFunction:
    """
    Class wrapping a function build with
    @see cl OnnxNumpyCompiler.

    .. versionadded:: 0.6
    """

    def __init__(self, compiler, rt, inputs, outputs,
                 n_optional, n_variables):
        if any(map(lambda n: not isinstance(n, Variable), inputs)):
            raise TypeError(  # pragma: no cover
                f"All inputs must be of type Variable: {inputs!r}.")
        if any(map(lambda n: not isinstance(n, Variable), outputs)):
            raise TypeError(  # pragma: no cover
                f"All outputs must be of type Variable: {outputs!r}.")
        self.compiler = compiler
        self.inputs = inputs
        self.outputs = outputs
        self.rt = rt
        self.n_optional = n_optional
        self.n_variables = n_variables
        if n_optional < 0:
            raise RuntimeError(  # pragma: no cover
                f"Wrong configuration, n_optional {n_optional!r} must be >= 0.")
        if n_optional >= len(inputs):
            raise RuntimeError(  # pragma: no cover
                "Wrong configuration, n_optional %r must be >= %r "
                "the number of inputs." % (n_optional, len(inputs)))

    def _check_(self, *args, **kwargs):
        if self.n_variables > 0:
            return
        if (len(args) < len(self.inputs) - self.n_optional or
                len(args) > len(self.inputs)):
            raise RuntimeError(  # pragma: no cover
                "Unexpected number of inputs %d. It should be in "
                "[%r, %r] len(args)=%d n_optional=%d n_variables=%d"
                "\nargs=%s\nkwargs=%s\ninputs=%s" % (
                    len(args), len(self.inputs) - self.n_optional,
                    len(args), self.n_optional, self.n_variables,
                    len(self.inputs), args, kwargs, self.inputs))


class OnnxNumpyFunctionOnnxInference(OnnxNumpyFunction):
    """
    Overwrites @see cl OnnxNumpyFunction to run an instance of
    @see cl OnnxInference.

    .. versionadded:: 0.6
    """

    def __call__(self, *args, **kwargs):
        self._check_(*args, **kwargs)
        inp = {k.name: a for k, a in zip(self.inputs, args)}
        out = self.rt.run(inp, **kwargs)
        if len(out) != len(self.outputs):
            raise RuntimeError(  # pragma: no cover
                "Unexpected number of outputs %d instead of %d." % (
                    len(out), len(self.outputs)))
        return tuple([out[o.name] for o in self.outputs])


class OnnxNumpyFunctionInferenceSession(OnnxNumpyFunction):
    """
    Overwrites @see cl OnnxNumpyFunction to run an instance of
    `InferenceSession` from :epkg:`onnxruntime`.

    .. versionadded:: 0.6
    """

    def __call__(self, *args, **kwargs):
        self._check_(*args, **kwargs)
        if len(kwargs) > 0:
            raise RuntimeError(  # pragma: no cover
                f"kwargs is not used but it is not empty: {kwargs!r}.")
        inp = {k.name: a for k, a in zip(self.inputs, args)}
        out = self.rt.run(None, inp)

        if len(out) != len(self.outputs):
            raise RuntimeError(  # pragma: no cover
                "Unexpected number of outputs %d instead of %d." % (
                    len(out), len(self.outputs)))
        return tuple(out)


class OnnxNumpyCompiler:
    """
    Implements a class which runs onnx graph.

    :param fct: a function with annotations which returns an ONNX graph,
        it can also be an ONNX graph.
    :param op_version: :epkg:`ONNX` opset to use, None
        for the latest one
    :param runtime: runtime to choose to execute the onnx graph,
        `python`, `onnxruntime`, `onnxruntime1`
    :param signature: used when the function is not annotated
    :param version: the same function can be instantiated with
        different type, this parameter is None or a numpy type
        if the signature allows multiple types, it must an instance
        of type @see cl FctVersion
    :param fctsig: function used to overwrite the fct signature
        in case this one is using `*args, **kwargs`

    .. versionadded:: 0.6
    """

    def __init__(self, fct, op_version=None, runtime=None, signature=None,
                 version=None, fctsig=None):
        if version is not None and not isinstance(version, FctVersion):
            raise TypeError(  # pragma: no cover
                "version must be of Type 'FctVersion' not %s - %s"
                "." % (type(version), version))
        self.fctsig = fctsig
        if op_version is None:
            from .. import __max_supported_opset__
            op_version = __max_supported_opset__
        if hasattr(fct, 'SerializeToString'):
            self.fct_ = None
            self.onnx_ = fct
        else:
            self.fct_ = fct
            if not inspect.isfunction(fct):
                raise TypeError(  # pragma: no cover
                    f"Unexpected type for fct={type(fct)!r}, it must be a function.")
            self.onnx_ = None
            self.onnx_ = self._to_onnx(
                op_version=op_version, signature=signature,
                version=version)
        self.runtime_ = self._build_runtime(
            op_version=op_version, runtime=runtime,
            signature=signature, version=version)
        ann = self._parse_annotation(signature=signature, version=version)
        inputs, outputs, kwargs, n_optional, n_variables = ann
        n_opt = 0 if signature is None else signature.n_optional
        args, kwargs2 = get_args_kwargs(self.fctsig or self.fct_, n_opt)
        self.meta_ = dict(op_version=op_version, runtime=runtime,
                          signature=signature, version=version,
                          inputs=inputs, outputs=outputs,
                          kwargs=kwargs, n_optional=n_optional,
                          n_variables=n_variables,
                          args=args, kwargs2=kwargs2,
                          annotations=self.fct_.__annotations__)

    def __getstate__(self):
        """
        Serializes everything but function `fct_`.
        Function `fct_` is used to build the onnx graph
        and is not needed anymore.
        """
        return dict(onnx_=self.onnx_, meta_=self.meta_)

    def __setstate__(self, state):
        """
        Restores serialized data.
        """
        for k, v in state.items():
            setattr(self, k, v)
        self.runtime_ = self._build_runtime(
            op_version=self.meta_['op_version'],
            runtime=self.meta_['runtime'],
            signature=self.meta_['signature'],
            version=self.meta_['version'])

    def __repr__(self):
        "usual"
        if self.fct_ is not None:
            return f"{self.__class__.__name__}({repr(self.fct_)})"
        if self.onnx_ is not None:
            return f"{self.__class__.__name__}({'... ONNX ... '})"
        raise NotImplementedError(  # pragma: no cover
            "fct_ and onnx_ are empty.")

    def _to_onnx_shape(self, shape):
        if shape is Any or shape is Ellipsis:
            shape = None
        elif isinstance(shape, tuple):
            shape = [None if s is Any or s is Ellipsis else s
                     for s in shape]
        else:
            raise RuntimeError(  # pragma: no cover
                f"Unexpected annotated shape {shape!r}.")
        return shape

    def _parse_annotation(self, signature, version):
        """
        Returns the annotations for function `fct_`.

        :param signature: needed if the annotation is missing,
            then version might be needed to specify which type
            to use if the signature allows many
        :param version: version inside the many signatures possible
        :return: *tuple(inputs, outputs, kwargs)*, each of them
            is a list of tuple with the name and the dtype,
            *kwargs* is the list of additional parameters
        """
        n_opt = 0 if signature is None else signature.n_optional
        if hasattr(self, 'meta_'):
            args, kwargs = self.meta_['args'], self.meta_['kwargs2']
        else:
            args, kwargs = get_args_kwargs(self.fctsig or self.fct_, n_opt)
        if version is not None:
            nv = len(version) - len(args) - n_opt
            if (signature is not None and not
                    signature.n_variables and nv > len(kwargs)):
                raise RuntimeError(  # pragma: no cover
                    "Mismatch (%d - %d - %d ? %d) between version=%r and kwargs=%r for "
                    "function %r, optional argument is %d, "
                    "signature=%r." % (
                        len(version), len(args), n_opt, len(kwargs),
                        version, kwargs, self.fct_,
                        signature.n_variables, signature))
            vvers = {} if version.kwargs is None else version.kwargs
            up = {}
            for k, v in zip(kwargs, vvers):
                up[k] = v
            kwargs = kwargs.copy()
            kwargs.update(up)

        for k, v in kwargs.items():
            if isinstance(v, (type, numpy.dtype)):
                raise RuntimeError(  # pragma: no cover
                    f"Unexpected value for argument {k!r}: {v!r} from {kwargs!r}.")

        if signature is not None:
            inputs, kwargs, outputs, n_optional, n_variables = (
                signature.get_inputs_outputs(args, kwargs, version))
            inputs = [Variable(i[0], i[1]) for i in inputs]
            outputs = [Variable(i[0], i[1]) for i in outputs]
            return inputs, outputs, kwargs, n_optional, n_variables

        def _possible_names():
            yield 'y'
            yield 'z'  # pragma: no cover
            yield 'o'  # pragma: no cover
            for i in range(0, 10000):  # pragma: no cover
                yield 'o%d' % i

        if hasattr(self, 'meta_'):
            annotations = self.meta_['annotations']
        else:
            annotations = self.fct_.__annotations__
        inputs = []
        outputs = []
        for a in args:
            if a == "op_version":
                continue
            if a not in annotations:
                raise RuntimeError(  # pragma: no cover
                    "Unable to find annotation for argument %r. "
                    "You should annotate the arguments and the results "
                    "or specify a signature." % a)
            ann = annotations[a]
            shape, dtype = ann.__args__
            shape = self._to_onnx_shape(shape)
            inputs.append(Variable(a, dtype, shape=shape))

        ret = annotations['return']
        names_in = set(inp.name for inp in inputs)

        if isinstance(ret, tuple):
            # multiple outputs
            names_none = set()
            for shape_dtype in ret:
                shape, dtype = shape_dtype.__args__
                shape = self._to_onnx_shape(shape)
                name_out = None
                for name in _possible_names():
                    if name not in names_in and name not in names_none:
                        name_out = name
                        break
                outputs.append(Variable(name_out, dtype, shape=shape))
                names_none.add(name_out)
            return (inputs, outputs, kwargs, 0,
                    signature.n_variables if signature is not None else False)

        # single outputs
        shape, dtype = ret.__args__
        shape = self._to_onnx_shape(shape)
        name_out = None
        for name in _possible_names():
            if name not in names_in:
                name_out = name
                break
        outputs.append(Variable(name_out, dtype, shape=shape))
        return (inputs, outputs, kwargs, 0,
                signature.n_variables if signature is not None else False)

    def _find_hidden_algebras(self, onx_var, onx_algebra):
        """
        Subgraph are using inputs not linked to the others nodes.
        This function retrieves them as they are stored in
        attributes `alg_hidden_var_`. The function looks into every
        node linked to the inputs and their predecessors.

        :param onx_var: @see cl OnnxVar
        :param onx_algebra: OnnxOperator
        :return: tuple(dictionary `{id(obj): (var, obj)}`,
            all instance of @see cl OnnxVarGraph)
        """
        keep_hidden = {}
        var_graphs = []
        stack = [onx_var]
        while len(stack) > 0:
            var = stack.pop()
            hidden = getattr(var, 'alg_hidden_var_', None)
            if hidden is not None:
                if any(map(lambda x: len(x) > 0,
                           var.alg_hidden_var_inputs.values())):
                    keep_hidden.update(hidden)
                    var_graphs.append(var)
            if hasattr(var, 'inputs'):
                for inp in var.inputs:
                    stack.append(inp)
        return keep_hidden, var_graphs

    def _to_onnx(self, op_version=None, signature=None, version=None):
        """
        Returns the onnx graph produced by function `fct_`.
        """
        if self.onnx_ is None and self.fct_ is not None:
            from .onnx_variable import OnnxVar
            logger.debug('OnnxNumpyCompiler._to_onnx(op_version=%r, '
                         'signature=%r, version=%r)',
                         op_version, signature, version)
            inputs, outputs, kwargs, n_optional, n_variables = (  # pylint: disable=W0612
                self._parse_annotation(
                    signature=signature, version=version))
            if ((signature is None or not signature.n_variables) and
                    isinstance(version, tuple) and
                    len(inputs) > len(version)):
                raise NotImplementedError(  # pragma: no cover
                    "Mismatch between additional parameters %r "
                    "(n_optional=%r) and version %r for function %r from %r."
                    "" % (kwargs, n_optional, version, self.fct_,
                          getattr(self.fct_, '__module__', None)))
            names_in = [oi.name for oi in inputs]
            names_out = [oi.name for oi in outputs]
            names_var = [OnnxVar(n, dtype=dt.dtype)
                         for n, dt in zip(names_in, inputs)]

            logger.debug('OnnxNumpyCompiler._to_onnx:names_in=%r', names_in)
            logger.debug('OnnxNumpyCompiler._to_onnx:names_out=%r', names_out)

            if 'op_version' in self.fct_.__code__.co_varnames:
                onx_var = None
                onx_algebra = self.fct_(
                    *names_in, op_version=op_version, **kwargs)
            else:
                onx_var = self.fct_(*names_var, **kwargs)
                if not hasattr(onx_var, 'to_algebra'):
                    raise TypeError(  # pragma: no cover
                        "The function %r to convert must return an instance of "
                        "OnnxVar but returns type %r." % (self.fct_, type(onx_var)))
                onx_algebra = onx_var.to_algebra(op_version=op_version)

            logger.debug('OnnxNumpyCompiler._to_onnx:onx_var=%r',
                         type(onx_var))
            logger.debug('OnnxNumpyCompiler._to_onnx:onx_algebra=%r',
                         type(onx_algebra))

            if not isinstance(onx_algebra, (OnnxOperator, OnnxOperatorTuple)):
                raise TypeError(  # pragma: no cover
                    "Unexpected type for onx_algebra %r "
                    "(It should be OnnxOperator or OnnxOperatorItem), "
                    "function is %r." % (type(onx_algebra), self.fct_))
            hidden_algebras, var_graphs = self._find_hidden_algebras(
                onx_var, onx_algebra)
            if len(hidden_algebras) > 0:
                logger.debug(  # pragma: no cover
                    'OnnxNumpyCompiler._to_onnx:len(hidden_algebras)=%r',
                    len(hidden_algebras))
                # print('----1', len(var_graphs))
                # for gr in var_graphs:
                #     print(type(gr), dir(gr))
                # print('----2', len(hidden_algebras))
                # for k, v in hidden_algebras.items():
                #     print("*", type(v.alg_), dir(v.alg_))
                #     #import pprint
                #     #pprint.pprint(dir(v.alg_))
                raise NotImplementedError(  # pragma: no cover
                    "Subgraphs only support constants (operator If, Loop, "
                    "Scan). hidden_algebras=%r var_graphs=%r" % (
                        hidden_algebras, var_graphs))

            if isinstance(onx_algebra, str):
                raise RuntimeError(  # pragma: no cover
                    f"Unexpected str type {onx_algebra!r}.")
            if isinstance(onx_algebra, tuple):
                raise NotImplementedError(  # pragma: no cover
                    "Not implemented when the function returns multiple results.")
            if hasattr(onx_algebra, 'to_onnx'):
                onx_algebra.output_names = [Variable(n) for n in names_out]
                onx = onx_algebra.to_onnx(
                    inputs=inputs, target_opset=op_version, outputs=outputs)
                # optimisation
                onx_optimized = onnx_optimisations(onx)
                self.onnx_ = onx_optimized

        if self.onnx_ is None:
            raise RuntimeError(  # pragma: no cover
                "Unable to get the ONNX graph (class %r, fct_=%r)" % (
                    type(self), self.fct_))
        return self.onnx_

    def to_onnx(self, **kwargs):
        """
        Returns the ONNX graph for the wrapped function.
        It takes additional arguments to distinguish between multiple graphs.
        This happens when a function needs to support multiple type.

        :return: ONNX graph
        """
        if len(kwargs) > 0:
            raise NotImplementedError(  # pragma: no cover
                "kwargs is not empty, this case is not implemented. "
                "kwargs=%r." % kwargs)
        if hasattr(self, 'onnx_'):
            return self.onnx_
        raise NotImplementedError(  # pragma: no cover
            "Attribute 'onnx_' is missing.")

    def _build_runtime(self, op_version=None, runtime=None,
                       signature=None, version=None):
        """
        Creates the runtime for the :epkg:`ONNX` graph.

        :param op_version: :epkg:`ONNX` opset to use, None
            for the latest one
        :param runtime: runtime to choose to execute the onnx graph,
            `python`, `onnxruntime`, `onnxruntime1`
        :param signature: used when the function is not annotated
        """
        onx = self._to_onnx(op_version=op_version, signature=signature,
                            version=version)
        inputs, outputs, _, n_optional, n_variables = self._parse_annotation(
            signature=signature, version=version)
        if runtime not in ('onnxruntime', 'onnxruntime-cuda'):
            from ..onnxrt import OnnxInference
            rt = OnnxInference(onx, runtime=runtime)
            self.rt_fct_ = OnnxNumpyFunctionOnnxInference(
                self, rt, inputs=inputs, outputs=outputs,
                n_optional=n_optional, n_variables=n_variables)
        else:
            from ..tools.ort_wrapper import InferenceSession
            rt = InferenceSession(onx.SerializeToString(), runtime=runtime)
            self.rt_fct_ = OnnxNumpyFunctionInferenceSession(
                self, rt, inputs=inputs, outputs=outputs,
                n_optional=n_optional, n_variables=n_variables)
        return self.rt_fct_

    def __call__(self, *args, **kwargs):
        """
        Executes the function and returns the results.

        :param args: arguments
        :return: results
        """
        res = self.rt_fct_(*args, **kwargs)
        if len(res) == 1:
            return res[0]
        return res



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from ._op import OpRun


class LabelEncoder(OpRun):

    atts = {'default_float': 0., 'default_int64': -1,
            'default_string': b'',
            'keys_floats': numpy.empty(0, dtype=numpy.float32),
            'keys_int64s': numpy.empty(0, dtype=numpy.int64),
            'keys_strings': numpy.empty(0, dtype=numpy.str_),
            'values_floats': numpy.empty(0, dtype=numpy.float32),
            'values_int64s': numpy.empty(0, dtype=numpy.int64),
            'values_strings': numpy.empty(0, dtype=numpy.str_),
            }

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       expected_attributes=LabelEncoder.atts,
                       **options)
        if len(self.keys_floats) > 0 and len(self.values_floats) > 0:
            self.classes_ = {k: v for k, v in zip(
                self.keys_floats, self.values_floats)}
            self.default_ = self.default_float
            self.dtype_ = numpy.float32
        elif len(self.keys_floats) > 0 and len(self.values_int64s) > 0:
            self.classes_ = {k: v for k, v in zip(
                self.keys_floats, self.values_int64s)}
            self.default_ = self.default_int64
            self.dtype_ = numpy.int64
        elif len(self.keys_int64s) > 0 and len(self.values_int64s) > 0:
            self.classes_ = {k: v for k, v in zip(
                self.keys_int64s, self.values_int64s)}
            self.default_ = self.default_int64
            self.dtype_ = numpy.int64
        elif len(self.keys_int64s) > 0 and len(self.values_floats) > 0:
            self.classes_ = {k: v for k, v in zip(
                self.keys_int64s, self.values_floats)}
            self.default_ = self.default_int64
            self.dtype_ = numpy.float32
        elif len(self.keys_strings) > 0 and len(self.values_floats) > 0:
            self.classes_ = {k.decode('utf-8'): v for k, v in zip(
                self.keys_strings, self.values_floats)}
            self.default_ = self.default_float
            self.dtype_ = numpy.float32
        elif len(self.keys_strings) > 0 and len(self.values_int64s) > 0:
            self.classes_ = {k.decode('utf-8'): v for k, v in zip(
                self.keys_strings, self.values_int64s)}
            self.default_ = self.default_int64
            self.dtype_ = numpy.int64
        elif len(self.keys_strings) > 0 and len(self.values_strings) > 0:
            self.classes_ = {
                k.decode('utf-8'): v.decode('utf-8') for k, v in zip(
                    self.keys_strings, self.values_strings)}
            self.default_ = self.default_string
            self.dtype_ = numpy.array(self.classes_.values).dtype
        elif len(self.keys_floats) > 0 and len(self.values_strings) > 0:
            self.classes_ = {k: v.decode('utf-8') for k, v in zip(
                self.keys_floats, self.values_strings)}
            self.default_ = self.default_string
            self.dtype_ = numpy.array(self.classes_.values).dtype
        elif len(self.keys_int64s) > 0 and len(self.values_strings) > 0:
            self.classes_ = {k: v.decode('utf-8') for k, v in zip(
                self.keys_int64s, self.values_strings)}
            self.default_ = self.default_string
            self.dtype_ = numpy.array(self.classes_.values).dtype
        elif hasattr(self, 'classes_strings'):
            raise RuntimeError(  # pragma: no cover
                "This runtime does not implement version 1 of "
                "operator LabelEncoder.")
        else:
            raise RuntimeError(
                f"No encoding was defined in {onnx_node}.")
        if len(self.classes_) == 0:
            raise RuntimeError(  # pragma: no cover
                "Empty classes for LabelEncoder, (onnx_node='{}')\n{}.".format(
                    self.onnx_node.name, onnx_node))

    def _run(self, x, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        if len(x.shape) > 1:
            x = numpy.squeeze(x)
        res = numpy.empty((x.shape[0], ), dtype=self.dtype_)
        for i in range(0, res.shape[0]):
            res[i] = self.classes_.get(x[i], self.default_)
        return (res, )
